from __future__ import annotations

from typing import Any

import jinja2

from staticsite.cmd.site import FeatureCommand
from staticsite.feature import Feature


class Hello(Feature):
    """
    Example custom feature for staticsite.

    It:
    * adds a hello metadata element to all pages that do not have it already
    * adds a hello() function to jinja2 that fetches the contents of their hello
      element
    """
    def __init__(self, *args: Any, **kw: Any):
        super().__init__(*args, **kw)
        self.j2_globals["hello"] = self.hello

    @jinja2.contextfunction
    def hello(self, context: jinja2.runtime.Context):
        # This function, called from a jinja2 template, has access to the
        # render context. staticsite puts 'page' in the render context,
        # pointing to the Page object being rendered.
        return context.page.meta.get("hello", "oddly, no hello here")

    def organize(self):
        # Add a 'hello' metadata element to all pages
        # This runs after 'taxonomies' and 'dirs', so it also annotates the
        # pages they generate
        for page in self.site.iter_pages(static=False):
            page.meta.setdefault("hello", "Hello " + page.meta.get("title", "page"))

    def add_site_commands(self, subparsers):
        # Add an example command as ssite site --cmd hello
        super().add_site_commands(subparsers)
        HelloCmd.add_subparser(subparsers)


class HelloCmd(FeatureCommand):
    "greet the user"

    NAME = "hello"

    def run(self) -> None:
        count = 0
        for page in self.site.iter_pages(static=False):
            count += 1
        print(f"Hello from {self.site.settings.SITE_NAME}!")
        print(f"I contain {count} pages.")


#  FEATURES dict defines which features are activated and with which name.
FEATURES = {
    "hello": Hello,
}



from __future__ import annotations

import logging
import sys
from collections import defaultdict
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast

from . import fields, file, fstree, site, toposort
from .node import Node
from .page import Page

if TYPE_CHECKING:
    import argparse

    from .archetypes import Archetype, Archetypes
    from .source_node import SourcePageNode

log = logging.getLogger("feature")


P = TypeVar("P", bound=Page)
V = TypeVar("V")
NT = TypeVar("NT", bound=type[Node])
PT = TypeVar("PT", bound=type[Page])


class TrackedField(fields.Field[P, V]):
    """
    Leat a feature track when values are set on this field
    """

    tracked_by: str

    def __init__(self, *, tracked_by: str | None = None, **kw: Any):
        """
        :arg tracked_by: name of a feature that is notified when field values
                         are ses
        """
        super().__init__(**kw)
        if tracked_by is not None:
            self.tracked_by = tracked_by
        self.track_feature_key: str

    def __set_name__(self, owner: type[P], name: str) -> None:
        super().__set_name__(owner, name)
        self.track_feature_key = f"_{self.name}_feature"

    def track(self, obj: P, value: Any) -> None:
        # We cannot store a reference to the Feature in the field, because it
        # would persist in the class definition, making it impossible to build
        # two sites one after the other.
        if (f := obj.__dict__.get(self.track_feature_key)) is None:
            feature = cast(PageTrackingMixin[P], obj.site.features[self.tracked_by])
            obj.__dict__[self.track_feature_key] = feature
        else:
            feature = cast(PageTrackingMixin[P], f)
        feature.track_field(self, obj, value)

    def __set__(self, obj: P, value: Any) -> None:
        self.track(obj, value)
        super().__set__(obj, value)


class Feature:
    """
    Base class for implementing a staticsite feature.

    It contains dependencies on other features, and hooks called in various
    stages of site processing.
    """

    # Name with which the feature class was loaded
    NAME: str

    # List names of features that should run after us.
    # The dependency order is taken into account when calling try_load_page and
    # finalize.
    RUN_BEFORE: list[str] = []

    # List names of features that should run before us.
    # The dependency order is taken into account when calling try_load_page and
    # finalize.
    RUN_AFTER: list[str] = []

    def __init__(self, name: str, site: site.Site):
        # Feature name
        self.name = name
        # Site object
        self.site = site
        # Feature-provided jinja2 globals
        self.j2_globals: dict[str, Callable[..., Any]] = {}
        # Feature-provided jinja2 filters
        self.j2_filters: dict[str, Callable[..., Any]] = {}
        # Feature-provided page mixins
        self.page_mixins: list[type[Page]] = []

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name})"

    def get_short_description(self) -> str:
        """
        Get a short description from this feature's docstring
        """
        if not self.__doc__:
            return ""
        else:
            return self.__doc__.lstrip().splitlines()[0].strip()

    def load_dir_meta(self, directory: fstree.Tree) -> dict[str, Any] | None:
        """
        Hook to load extra directory metadata for the given sitedir.

        sitedir will be already populated with directory contents, but not yet
        loaded.
        """
        # Do nothing by default
        return None

    def load_dir(
        self,
        node: SourcePageNode,
        directory: fstree.Tree,
        files: dict[str, tuple[dict[str, Any], file.File]],
    ) -> list[Page]:
        """
        Load pages from the given Dir.

        Remove from dir the filenames that have been loaded.

        Return the list of loaded pages.
        """
        return []

    def try_load_archetype(
        self, archetypes: Archetypes, relpath: str, name: str
    ) -> Archetype | None:
        """
        Try loading an archetype page.

        Returns None if this path is not handled by this feature
        """
        return None

    def organize(self) -> None:
        """
        Hook called after all the pages have been loaded
        """

    def generate(self) -> None:
        """
        Hook called after loaded pages have been organized, to autogenerate new
        pages
        """

    def crossreference(self) -> None:
        """
        Hook called after autogenerated pages have been added and the site
        structure is stable
        """

    def add_site_commands(self, subparsers: argparse._SubParsersAction[Any]) -> None:
        """
        Add commands to `ssite site --cmd …` command line parser
        """

    def get_node_bases(self) -> Sequence[type[Node]]:
        """
        Return a sequence of base classes to add when instantiating Node types
        """
        return ()

    def get_page_bases(self, page_cls: type[Page]) -> Sequence[type[Page]]:
        """
        Return a sequence of base classes to add when instantiating the given
        page type
        """
        return ()

    def get_used_page_types(self) -> list[type[Page]]:
        """
        Return a list of the page types used by this feature
        """
        return []

    def get_footprint(self) -> dict[str, Any]:
        """
        Return information that will be cached until the next build to allow
        incremental updates
        """
        return {}

    def set_previous_footprint(self, footprint: dict[str, Any]) -> None:
        """
        Notify the feature of a cached footprint from a previous build
        """


class Features:
    def __init__(self, site: site.Site):
        self.site = site

        # Registry of feature classes by name, built during loading
        self.feature_classes: dict[str, type[Feature]] = {}

        # Feature implementation registry
        self.features: dict[str, Feature] = {}

        # Cached final node classes indexed by original node class
        self.node_classes: dict[type[Node], type[Node]] = {}

        # Cached final page classes indexed by original page class
        self.page_classes: dict[type[Page], type[Page]] = {}

        # Features sorted by topological order
        self.sorted: list[Feature] | None = None

    def get_node_class(self, cls: NT) -> NT:
        """
        Given a base Node class, return its version with added mixins
        """
        if final := self.node_classes.get(cls):
            return cast(NT, final)
        bases: list[type[Node]] = []
        for feature in self.ordered():
            bases.extend(feature.get_node_bases())
        final = type(
            cls.__name__,
            tuple(bases) + (cls,),
            {"__doc__": cls.__doc__, "__module__": cls.__module__},
        )
        self.node_classes[cls] = final
        return cast(NT, final)

    def get_page_class(self, cls: PT) -> PT:
        """
        Given a base Page class, return its version with added mixins
        """
        if final := self.page_classes.get(cls):
            return cast(PT, final)
        bases: list[type[Page]] = []
        for feature in self.ordered():
            bases.extend(feature.get_page_bases(cls))
        final = type(
            cls.__name__,
            tuple(bases) + (cls,),
            {"__doc__": cls.__doc__, "__module__": cls.__module__},
        )
        self.page_classes[cls] = final
        return cast(PT, final)

    def ordered(self) -> list[Feature]:
        if self.sorted is None:
            raise RuntimeError(
                "Features.ordered called before features have finished loading"
            )
        return self.sorted

    def __getitem__(self, key: str, /) -> Feature:
        return self.features[key]

    def get(self, key: str, default: Feature | None = None) -> Feature | None:
        return self.features.get(key, default)

    def _sort_features(self, features: dict[str, type[Feature]]) -> list[type[Feature]]:
        graph: dict[str, set[str]] = defaultdict(set)

        # Add feature dependencies
        for feature in features.values():
            # Make sure that each feature is in the graph
            graph[feature.NAME] = set()

            for name in feature.RUN_AFTER:
                if name not in features and name not in graph:
                    log.warning(
                        "feature %s: ignoring RUN_AFTER relation on %s which is not available",
                        feature,
                        name,
                    )
                    continue
                graph[feature.NAME].add(name)

            for name in feature.RUN_BEFORE:
                if name not in features and name not in graph:
                    log.warning(
                        "feature %s: ignoring RUN_BEFORE relation on %s which is not available",
                        feature,
                        name,
                    )
                    continue
                graph[name].add(feature.NAME)

        # Build the sorted list of features
        sorted_names = toposort.sort(graph)
        log.debug("Feature run order: %r", sorted_names)
        sorted_features = []
        for name in sorted_names:
            f = features.get(name)
            # Skip names that are not features, like well-known synchronization
            # points
            if f is None:
                continue
            sorted_features.append(f)
        return sorted_features

    def commit(self) -> None:
        """
        Finalize feature loading, instantiating and initializing all the
        features that have been collected.
        """
        self.sorted = []

        # Instantiate the feature classes in dependency order
        for cls in self._sort_features(self.feature_classes):
            if cls.NAME in self.features:
                continue
            feature = cls(cls.NAME, self.site)
            self.features[cls.NAME] = feature
            self.sorted.append(feature)

        log.debug("sorted feature list: %r", [x.name for x in self.sorted])

    def load_default_features(self) -> None:
        """
        Load features packaged with staticsite
        """
        from . import features

        self.load_feature_dir(features.__path__)

    def load_feature_dir(
        self, paths: Sequence[str], namespace: str = "staticsite.features"
    ) -> None:
        """
        Load all features found in the given directory.

        Feature classes are instantiate in dependency order.
        """
        import importlib
        import pkgutil

        for module_finder, name, ispkg in pkgutil.iter_modules(paths):
            full_name = namespace + "." + name
            mod = sys.modules.get(full_name)
            if not mod:
                try:
                    spec = module_finder.find_spec(name)
                    if spec is None:
                        log.error("spec not found for module %r", name)
                        continue
                    mod = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(mod)
                except Exception:
                    log.exception("%r: failed to load feature module", name)
                    continue
                sys.modules[full_name] = mod

            features = getattr(mod, "FEATURES", None)
            if features is None:
                log.warning("%r: feature module did not define a FEATURES dict", name)
                continue

            # Register features with site
            for name, cls in features.items():
                cls.NAME = name
                old = self.feature_classes.get(name)
                if old is not None:
                    # Allows replacing features: see #28
                    log.info("%s: replacing feature %s with %s", name, old, cls)
                self.feature_classes[name] = cls


PG = TypeVar("PG", bound="fields.FieldContainer")


class PageTrackingMixin(Generic[PG]):
    """
    Basic page tracking functionality for features that need it
    """

    def __init__(self, *args: Any, **kw: Any) -> None:
        super().__init__(*args, **kw)
        # Collect pages notified by track_field, regardless of field name
        self.tracked_pages: set[PG] = set()

    def track_field(self, field: fields.Field[PG, Any], obj: PG, value: Any) -> None:
        self.tracked_pages.add(obj)



from __future__ import annotations

import logging
from collections.abc import Callable, Iterator
from typing import TYPE_CHECKING, Any, NamedTuple

import jinja2
import markupsafe

from staticsite.feature import Feature
from staticsite.page import ChangeExtent, Page, SourcePage, TemplatePage
from staticsite.page_filter import compile_page_match
from staticsite.utils import front_matter

if TYPE_CHECKING:
    from staticsite import file, fstree
    from staticsite.source_node import SourcePageNode

log = logging.getLogger("jinja2")


def load_front_matter(template: jinja2.Template) -> dict[str, Any]:
    """
    Load front matter from a jinja2 template
    """
    # If the page has a front_matter block, render it to get the front matter
    front_matter_block = template.blocks.get("front_matter")
    if front_matter_block:
        fm = "".join(front_matter_block(template.new_context()))
        fmt, meta = front_matter.read_string(fm)
    else:
        meta = {}

    # If the metadata do not have a title, try rendering the title block
    if "title" not in meta:
        title_block = template.blocks.get("title")
        if title_block:
            try:
                title = "".join(title_block(template.new_context())).strip()
            except jinja2.exceptions.TemplateError as e:
                log.warning(
                    "%s: cannot extract title from {%% block title %%}: %s",
                    template.name,
                    e,
                )
                title = None
            if title:
                meta["title"] = title

    return meta


class J2Pages(Feature):
    """
    Render jinja2 templates from the contents directory.

    See doc/reference/templates.md for details.
    """

    def get_used_page_types(self) -> list[type[Page]]:
        return [J2Page]

    def load_dir_meta(self, directory: fstree.Tree) -> dict[str, Any] | None:
        # Load front matter from index.html
        if (index := directory.files.get("index.html")) is None:
            return None

        try:
            template = self.site.theme.jinja2.get_template("content:" + index.relpath)
        except Exception:
            log.exception("%s: cannot load template", index.relpath)
        else:
            if front_matter := load_front_matter(template):
                return front_matter
        return None

    def load_dir(
        self,
        node: SourcePageNode,
        directory: fstree.Tree,
        files: dict[str, tuple[dict[str, Any], file.File]],
    ) -> list[Page]:
        # Precompile JINJA2_PAGES patterns
        want_patterns = [compile_page_match(p) for p in self.site.settings.JINJA2_PAGES]

        taken: list[str] = []
        pages: list[Page] = []
        for fname, (kwargs, src) in files.items():
            # Skip files that do not match JINJA2_PAGES
            for pattern in want_patterns:
                if pattern.match(fname):
                    break
            else:
                continue

            try:
                template = self.site.theme.jinja2.get_template("content:" + src.relpath)
            except Exception:
                log.exception("%s: cannot load template", src.relpath)
                continue

            front_matter = load_front_matter(template)
            if front_matter:
                kwargs.update(front_matter)

            kwargs["page_cls"] = J2Page
            kwargs["src"] = src
            kwargs["template"] = template

            # TODO: Is this replace still needed? Do we still
            # support the .j2 trick?
            fname = fname.replace(".j2", "")

            if fname == "index.html":
                page = node.create_source_page_as_index(**kwargs)
            else:
                page = node.create_source_page_as_file(dst=fname, **kwargs)

            if page is not None:
                pages.append(page)

            taken.append(fname)

        for fname in taken:
            del files[fname]

        return pages


class Block(NamedTuple):
    name: str
    block: Callable[[jinja2.runtime.Context], Iterator[str]]


class RenderPartialTemplateMixin(TemplatePage):
    def _find_block(self, *names: str) -> Block | None:
        for name in names:
            block = self.page_template.blocks.get("page_content")
            block_name = "page_content"
            if block is not None:
                return Block(block_name, block)
        log.warning(
            "%s: `page_content` and `content` not found in template %s",
            self,
            self.page_template.name,
        )
        return None

    @jinja2.pass_context
    def html_body(self, context: jinja2.runtime.Context, **kw: Any) -> str:
        if (block := self._find_block("page_content", "content")) is None:
            return ""
        return self.render_template_block(block, context, render_style="body")

    @jinja2.pass_context
    def html_inline(self, context: jinja2.runtime.Context, **kw: Any) -> str:
        if (block := self._find_block("page_content", "content")) is None:
            return ""
        return self.render_template_block(block, context, render_style="inline")

    @jinja2.pass_context
    def html_feed(self, context: jinja2.runtime.Context, **kw: Any) -> str:
        if (block := self._find_block("page_content", "content")) is None:
            return ""
        return self.render_template_block(block, context, render_style="feed")

    def render_template_block(
        self, block: Block, context: jinja2.runtime.Context, **kw: Any
    ) -> str:
        render_stack = list(context.get("render_stack", ()))

        render_style = kw.get("render_style")

        if (self, render_style) in render_stack:
            render_stack.append((self, render_style))
            render_stack_path = " → ".join(f"{p}:{s}" for p, s in render_stack)
            raise RuntimeError(f"{self}: render loop detected: {render_stack_path}")

        render_stack.append((self, render_style))
        kw["render_stack"] = render_stack
        kw["page"] = self

        try:
            return markupsafe.Markup(
                "".join(
                    block.block(
                        self.page_template.new_context(context, shared=True, locals=kw)
                    )
                )
            )
        except jinja2.TemplateError as e:
            log.error(
                "%s: %s: failed to render block %s: %s",
                self,
                self.page_template.filename,
                block.name,
                e,
            )
            log.debug(
                "%s: %s: failed to render block %s: %s",
                self,
                self.page_template.filename,
                block.name,
                e,
                exc_info=True,
            )
            # TODO: return a "render error" page? But that risks silent errors
            return ""


class J2Page(RenderPartialTemplateMixin, TemplatePage, SourcePage):
    """
    Jinja2 pages

    You can put [jinja2 templates](templates.md) in your site contents, and they
    will be rendered as site pages.

    This can be used to generate complex index pages, blog front pages, and
    anything else [Jinja2](http://jinja.pocoo.org/) is able to generate.

    You can set `JINJA2_PAGES` in the [site settings](settings.md) to a list of
    patterns (globs or regexps as in [page filters](page-filter.md)) matching file
    names to consider jinja2 templates by default. It defaults to
    `["*.html", "*.j2.*"]`.

    Any file with `.j2.` in their file name will be rendered as a template,
    stripping `.j2.` in the destination file name.

    For example, `dir/robots.j2.txt` will become `dir/robots.txt` when the site is
    built.

    ## Front matter

    If a page defines a jinja2 block called `front_matter`, the block is rendered
    and parsed as front matter.

    **Note**: [jinja2 renders all contents it finds before
    `{% extends %}`](https://jinja.palletsprojects.com/en/2.10.x/templates/#child-template).
    To prevent your front matter from ending up in the rendered HTML, place the
    `front_matter` block after the `{% extends %}` directive, or manage your front
    matter from [`.staticfile` directory metadata](content.md).

    If you want to use `template_*` entries, you can wrap the front matter around
    `{% raw %}` to prevent jinja2 from rendering their contents as part of the rest
    of the template.
    """

    TYPE = "jinja2"

    def __init__(self, *args: Any, template: jinja2.Template, **kw: Any):
        # Indexed by default
        kw.setdefault("indexed", True)
        super().__init__(*args, **kw)

        self.template = template

    def _compute_change_extent(self) -> ChangeExtent:
        # TODO: with some more infrastructure, we can track what pages
        # contributed the links, and compute something better.
        #
        # To track this we need to study the template system to see if there is
        # a way to find what page filter expressions are used, or what pages
        # are looked up, during last render.
        return ChangeExtent.ALL


FEATURES = {
    "jinja2": J2Pages,
}



from __future__ import annotations

import fnmatch
import os
import re
from collections.abc import Callable, Generator, Sequence
from typing import TYPE_CHECKING, Any, cast

from . import site
from .page import Page

if TYPE_CHECKING:
    from .node import Node


def compile_page_match(pattern: str | re.Pattern[str]) -> re.Pattern[str]:
    """
    Return a compiled re.Pattern from a glob or regular expression.

    :arg pattern:
      * if it's a re.Pattern instance, it is returned as is
      * if it starts with ``^`` or ends with ``$``, it is compiled as a regular
        expression
      * otherwise, it is considered a glob expression, and fnmatch.translate()
        is used to convert it to a regular expression, then compiled
    """
    if isinstance(pattern, re.Pattern):
        return pattern
    if pattern and (pattern[0] == "^" or pattern[-1] == "$"):
        return re.compile(pattern)
    return re.compile(fnmatch.translate(pattern))


def sort_args(
    sort: str | None,
) -> tuple[str | None, bool, Callable[[Page], Any] | None]:
    """
    Parse a page sort string, returning a tuple of:
    * which page metadata is used for sorting, or None if sorting is not based
      on metadata
    * a bool, set to True if sort order is reversed
    * a key function for the sorting
    """
    if sort is None:
        return None, False, None

    # Process the '-'
    if sort.startswith("-"):
        reverse = True
        sort = sort[1:]
    else:
        reverse = False

    # Add a sort key function
    if sort == "url":

        def key(page: Page) -> Any:
            return page.site_path

        return None, reverse, key
    else:

        def key(page: Page) -> Any:
            return getattr(page, cast(str, sort), None)

        return sort, reverse, key


class PageFilter:
    """
    Engine for selecting pages in the site
    """

    def __init__(
        self,
        site: site.Site,
        *,
        path: str | None = None,
        limit: int | None = None,
        sort: str | None = None,
        root: Node | None = None,
        allow: Sequence[Page] | None = None,
        **kw: str,
    ):
        self.site = site
        self.root = root or site.root

        self.re_path: re.Pattern[str] | None
        if path is not None:
            self.re_path = compile_page_match(path)
        else:
            self.re_path = None

        self.sort_meta, self.sort_reverse, self.sort_key = sort_args(sort)

        self.taxonomy_filters: list[tuple[str, frozenset[str]]] = []
        if (taxonomy_feature := self.site.features.get("taxonomy")) is not None:
            from staticsite.features.taxonomy import TaxonomyFeature

            for taxonomy in cast(TaxonomyFeature, taxonomy_feature).taxonomies.values():
                t_filter = kw.get(taxonomy.name)
                if t_filter is None:
                    continue
                self.taxonomy_filters.append((taxonomy.name, frozenset(t_filter)))

        self.limit = limit

        self.allow = allow

        # print(f"PageFilter({path=!r}, {self.root.path=!r}, {self.re_path=!r}, {self.taxonomy_filters=}")

    def filter(self) -> list[Page]:
        # print("PageFilter.filter")
        pages = []

        for page in self._filter(self.root or self.site.root, relpath=""):
            pages.append(page)

        if self.sort_key is not None:
            pages.sort(key=self.sort_key, reverse=self.sort_reverse)

        if self.limit is not None:
            pages = pages[: self.limit]

        return pages

    def _filter(self, root: Node, relpath: str) -> Generator[Page, None, None]:
        """
        :arg:relpath: path of the page relative to the root node of the search
        """
        for name, page in root.build_pages.items():
            # print(f"_filter {page=!r} indexed={page.meta['indexed']}")
            if not page.indexed:
                continue
            if self.allow is not None and page not in self.allow:
                continue
            if self.sort_meta is not None and self.sort_meta not in page.meta:
                continue

            # Taxonomy_filters
            fail_taxonomies = False
            for name, t_filter in self.taxonomy_filters:
                page_tags = frozenset(t.name for t in getattr(page, name, ()))
                if not t_filter.issubset(page_tags):
                    fail_taxonomies = True
            if fail_taxonomies:
                continue

            # print(f"_filter {page=!r} {relpath=!r} {page.dst}")

            if self.re_path is not None:
                if self.re_path.match(os.path.join(relpath, name)):
                    pass
                elif page.source_name and self.re_path.match(
                    os.path.join(relpath, page.source_name)
                ):
                    pass
                else:
                    # print(f"  {page=!r} {page.src=} did not match")
                    continue

            yield page

        if root.sub:
            for node in root.sub.values():
                yield from self._filter(node, relpath=os.path.join(relpath, node.name))



#!/usr/bin/python3
import argparse
import importlib
import logging
import sys

from staticsite import __version__

log = logging.getLogger("ssite")


def main():
    parser = argparse.ArgumentParser(prog="ssite", description="Static site generator.")
    parser.add_argument(
        "--version", "-V", action="version", version=f"%(prog)s {__version__}"
    )
    subparsers = parser.add_subparsers(help="sub-command help", dest="command")

    modules = (
        "check",
        "dump",
        "dump_meta",
        "build",
        "serve",
        "new",
        "edit",
        "site",
        "meta",
        "shell",
    )
    for name in modules:
        importlib.import_module("staticsite.cmd." + name)

    from staticsite.cmd.command import COMMANDS

    for c in COMMANDS:
        c.add_subparser(subparsers)

    args = parser.parse_args()
    if args.command is None:
        parser.print_help()
    else:
        handler = args.command(args)
        return handler.run()


def run_main():
    from staticsite.cmd.command import Fail, Success

    try:
        sys.exit(main())
    except Fail as e:
        print(e, file=sys.stderr)
        sys.exit(1)
    except Success:
        pass


if __name__ == "__main__":
    run_main()



from __future__ import annotations

import json
import logging
import re
from collections.abc import Generator, Iterable
from typing import IO, Any

from . import yaml_codec

log = logging.getLogger("utils")


def write(meta: dict[str, Any], style: str = "toml") -> str:
    if style == "json":
        return json.dumps(meta, indent=4, sort_keys=True)
    elif style == "toml":
        import toml

        return "+++\n" + toml.dumps(meta) + "+++\n"
    elif style == "yaml":
        return yaml_codec.dumps(meta) + "---\n"
    return ""


def read_markdown_partial(
    fd: IO[bytes],
) -> tuple[str | None, dict[str, Any], Iterable[str]]:
    """
    Parse lines front matter from a markdown file header.

    Stop reading at the end of the front matter.

    Returns the format of the front matter read, and parsed front matter.
    """
    buf = bytearray()
    line = fd.readline()
    head = line.strip()
    if head == b"{":
        # JSON, end at }
        buf += line
        while True:
            line = fd.readline()
            if not line:
                raise ValueError("unterminated json front matter")
            buf += line
            if line.rstrip() == b"}":
                return (
                    "json",
                    json.loads(buf.decode()),
                    (x.rstrip().decode() for x in fd),
                )
    elif head == b"---":
        # YAML, end at ---
        buf += line
        while True:
            line = fd.readline()
            if not line or line.rstrip() == b"---":
                return (
                    "yaml",
                    yaml_codec.loads(buf.decode()),
                    (x.rstrip().decode() for x in fd),
                )
            buf += line
    elif head == b"+++":
        # TOML, end at +++
        while True:
            line = fd.readline()
            if not line or line.rstrip() == b"+++":
                import toml

                return (
                    "toml",
                    toml.loads(buf.decode()),
                    (x.rstrip().decode() for x in fd),
                )
            buf += line
    elif head.startswith(b"```"):
        # Markdown verbatim block, end at ```, format follows head
        while True:
            line = fd.readline()
            if not line or line.rstrip() == b"```":
                break
            buf += line
        if head[3:] == b"yaml":
            return (
                "yaml",
                yaml_codec.loads(buf.decode()),
                (x.rstrip().decode() for x in fd),
            )
        elif head[3:] == b"toml":
            import toml

            return "toml", toml.loads(buf.decode()), (x.rstrip().decode() for x in fd)
        elif head[3:] == b"json":
            return "json", json.loads(buf.decode()), (x.rstrip().decode() for x in fd)
        else:
            return (
                "yaml",
                yaml_codec.loads(buf.decode()),
                (x.rstrip().decode() for x in fd),
            )
    else:
        # No front matter found
        def iter_body() -> Generator[str, None, None]:
            yield line.rstrip().decode()
            for x in fd:
                yield x.rstrip().decode()

        return None, {}, iter_body()


re_toml = re.compile(r"^\+\+\+[ \t]*\n(.+?)\n\+\+\+[ \t]*\n$", re.DOTALL)


def read_whole(fd: IO[str]) -> tuple[str, dict[str, Any]]:
    """
    Parse lines front matter from a file header.

    Read the entire file

    Returns the format of the front matter read, and parsed front matter
    """
    content = fd.read()
    if content.startswith("{"):
        return "json", (val if (val := json.loads(content)) is not None else {})

    mo = re_toml.match(content)
    if mo:
        import toml

        return "toml", (val if (val := toml.loads(mo.group(1))) is not None else {})

    return "yaml", (val if (val := yaml_codec.loads(content)) is not None else {})


def read_string(content: str) -> tuple[str, dict[str, Any]]:
    """
    Parse lines front matter from a file header.

    Read data from a string.

    Returns the format of the front matter read, and parsed front matter
    """
    if content.startswith("{"):
        return "json", json.loads(content)

    mo = re_toml.match(content)
    if mo:
        import toml

        return "toml", toml.loads(mo.group(1))

    return "yaml", yaml_codec.loads(content)



from __future__ import annotations

import argparse
import logging
import shlex
import subprocess
import tempfile
from typing import Any

from ..cache import DisabledCache
from ..utils import images
from ..utils import yaml_codec as yaml
from .command import Command, Fail, Success, register

log = logging.getLogger("meta")


@register
class Meta(Command):
    """
    Edit metadata for a file
    """

    def __init__(self, *args: Any, **kw: Any):
        super().__init__(*args, **kw)
        self.scanner = images.ImageScanner(DisabledCache("disabled"))

    def edit(self, fname: str) -> subprocess.CompletedProcess[bytes]:
        """
        Run the editor on this file
        """
        settings_dict = self.settings.as_dict()
        cmd = [
            x.format(name=fname, **settings_dict) for x in self.settings.EDIT_COMMAND
        ]
        try:
            res = subprocess.run(cmd, check=True)
        except subprocess.CalledProcessError as e:
            raise Fail(
                "Editor command {} exited with error {}".format(
                    " ".join(shlex.quote(x) for x in cmd), e.returncode
                )
            )
        return res

    def edit_meta(self, meta: dict[str, Any]) -> dict[str, Any]:
        """
        Edit the given metadata in an editor
        """
        with tempfile.NamedTemporaryFile(suffix=".yaml", mode="wt") as fd:
            # FIXME: ignoring type, as yaml.dump should declare IO[str] instead of TextIO
            yaml.dump(meta, fd)
            fd.flush()
            self.edit(fd.name)
            with open(fd.name) as newfd:
                if not isinstance((res := yaml.load(newfd)), dict):
                    raise RuntimeError("YAML did not parse as a dict")
                return res

    def save_changes(self, old_meta: dict[str, Any], new_meta: dict[str, Any]) -> None:
        # Compute and store the changes
        changed = {}
        removed = []
        for key, orig in old_meta.items():
            if key not in new_meta:
                log.info("%s: removed %s", self.args.file, key)
                removed.append(key)
            elif new_meta[key] != orig:
                log.info("%s: updated %s=%r", self.args.file, key, new_meta[key])
                changed[key] = new_meta[key]

        for key in new_meta.keys() - old_meta.keys():
            log.info("%s: added %s=%r", self.args.file, key, new_meta[key])
            changed[key] = new_meta[key]

        if not changed and not removed:
            raise Success()

        if not self.scanner.edit_meta_exiftool(self.args.file, changed, removed):
            raise Fail("Failed to store metadatä changes")

    def run(self) -> None:
        # TODO: Build a Site if possible
        # TODO: or load settings from a settings.py if one can be found in reasonable places

        # Read file metadata
        meta = self.scanner.scan_file(self.args.file)

        # Filter the keys that we don't need to edit
        meta.pop("width", None)
        meta.pop("height", None)
        meta.pop("lat", None)
        meta.pop("lon", None)

        # Set the keys that are relevant and might not be there to default values
        meta.setdefault("title", "")
        meta.setdefault("author", "")
        meta.setdefault("copyright", "")

        # Edit the metadata
        new_meta = self.edit_meta(meta)

        self.save_changes(meta, new_meta)

    @classmethod
    def add_subparser(
        cls, subparsers: argparse._SubParsersAction[Any]
    ) -> argparse.ArgumentParser:
        parser = super().add_subparser(subparsers)
        parser.add_argument("file", help="edit the metadata of this file")
        return parser



from __future__ import annotations

import contextlib
import datetime
import logging
import os
import shutil
import stat
import tempfile
from collections.abc import Sequence
from contextlib import ExitStack, contextmanager
from functools import cached_property
from typing import TYPE_CHECKING, Any, Union
from unittest import TestCase, mock

import pytz

from staticsite.cmd.build import Builder
from staticsite.file import File
from staticsite.settings import Settings
from staticsite.site import Site
from staticsite.utils import front_matter

if TYPE_CHECKING:
    from staticsite.page import Page

MockFiles = dict[str, Union[str, bytes, dict]]

project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


class MockSiteBase:
    """
    Common code for different ways of setting up mock sites
    """

    def __init__(
        self, auto_load_site: bool = True, settings: dict[str, Any] | None = None
    ):
        # Set to False if you only want to populate the workdir
        self.auto_load_site = auto_load_site
        self.stack = contextlib.ExitStack()
        self.root: str | None = None
        self.test_case: TestCase | None = None
        self.settings = Settings()

        self.settings.CACHE_REBUILDS = False
        self.settings.THEME_PATHS = [os.path.join(project_root, "themes")]
        self.settings.TIMEZONE = "Europe/Rome"

        if settings is not None:
            for k, v in settings.items():
                setattr(self.settings, k, v)

        # Timestamp used for mock files and site generation time
        # date +%s --date="2019-06-01 12:30"
        self.generation_time: int | None = 1559385000

        # Mock stat to return this mtime for files scanned during site load
        # date +%s --date="2019-06-01 12:30"
        self.mock_file_mtime: int | None = 1559385000

        self.root = self.stack.enter_context(tempfile.TemporaryDirectory())
        self.builder: Builder | None = None

    def populate_workdir(self):
        raise NotImplementedError(
            f"{self.__class__.__name__}.populate_workdir not implemented"
        )

    def create_site(self) -> Site:
        return Site(
            self.settings,
            generation_time=(
                datetime.datetime.fromtimestamp(self.generation_time, pytz.utc)
                if self.generation_time
                else None
            ),
        )

    @cached_property
    def site(self) -> Site:
        return self.create_site()

    @cached_property
    def build_root(self) -> str:
        return self.stack.enter_context(tempfile.TemporaryDirectory())

    def load_site(self, until=Site.LOAD_STEP_ALL):
        if self.mock_file_mtime:
            overrides = {"st_mtime": self.mock_file_mtime}
        else:
            overrides = None
        with mock_file_stat(overrides):
            self.site.load(until=until)

    def build_site(self):
        """
        Build the site in a temporary directory.

        The build directory will be available as self.build_root.

        The Builder object will be available as self.builder
        """
        self.site.settings.OUTPUT = self.build_root
        self.builder = Builder(self.site)
        self.builder.write()

    def page(self, *paths: tuple[str]) -> tuple[Page]:
        """
        Ensure the site has the given page, by path, and return it
        """
        res: list[Page] = []
        for path in paths:
            page = self.site.root.resolve_path(path)
            if page is None:
                self.test_case.fail(f"Page {path!r} not found in site")
            res.append(page)
        if len(res) == 1:
            return res[0]
        else:
            return tuple(res)

    def assertPagePaths(self, paths: Sequence[str]):
        """
        Check that the list of pages in the site matches the given paths
        """
        self.test_case.assertCountEqual(
            [p.site_path for p in self.site.iter_pages(static=False)], paths
        )

    def _find_page_by_relpath(self, relpath: str) -> Page | None:
        for page in self.site.iter_pages():
            if page.src.relpath == relpath:
                return page
        return None

    def assertBuilt(
        self,
        srcpath: str | None,
        sitepath: str,
        dstpath: str,
        sample: str | bytes | None = None,
    ):
        """
        Check that page at `srcpath` is in the site at `sitepath` and rendered in `dstpath`.

        `srcpath` can be None for autogenerated pages.

        Optionally check that the rendered content contains `sample`
        """
        if not self.builder:
            self.build_site()

        page = self.page(sitepath)
        if srcpath is not None:
            if page is None:
                found = self._find_page_by_relpath(srcpath)
                if found is None:
                    self.test_case.fail(
                        "{sitepath!r} not found in site, and {srcpath} not found in website"
                    )
                else:
                    self.test_case.fail(
                        "{sitepath!r} not found in site, and {srcpath} found at {found} instead"
                    )
            elif page.src.relpath != srcpath:
                found = self._find_page_by_relpath(srcpath)
                if found is None:
                    self.test_case.fail(
                        f"{sitepath!r} found in site, but for {page.src.relpath} instead of {srcpath}."
                        f" {srcpath} not found in website"
                    )
                else:
                    self.test_case.fail(
                        f"{sitepath!r} found in site, but for {page.src.relpath} instead of {srcpath}."
                        f" {srcpath} is instead at {found}"
                    )

        rendered = self.builder.build_log.get(dstpath)
        if rendered is None:
            for path, pg in self.builder.build_log.items():
                if pg == page:
                    self.test_case.fail(
                        f"{dstpath!r} not found in render log; {sitepath!r} was rendered as {path!r} instead"
                    )
                    break
            else:
                self.test_case.fail(f"{dstpath!r} not found in render log")

        if rendered != page:
            for path, pg in self.builder.build_log.items():
                if pg == page:
                    self.test_case.fail(
                        f"{dstpath!r} rendered {rendered!r} instead of {page!r}."
                        " {sitepath!r} was rendered as {path!r} instead"
                    )
                    break
            else:
                self.test_case.fail(
                    f"{dstpath!r} rendered {rendered!r} instead of {page!r}"
                )

        if os.path.isdir(os.path.join(self.build_root, dstpath)):
            self.test_case.fail(f"{dstpath!r} rendered as a directory")

        if sample is not None:
            if isinstance(sample, bytes):
                args = {"mode": "rb"}
            else:
                args = {"mode": "rt", "encoding": "utf-8"}
            with open(os.path.join(self.build_root, dstpath), **args) as fd:
                if sample not in (body := fd.read()):
                    self.test_case.fail(
                        f"{dstpath!r} does not contain {sample!r}. Renrered contents: {body!r}"
                    )

    def __enter__(self) -> MockSite:
        self.populate_workdir()
        if self.auto_load_site:
            self.load_site()
        return self

    def __exit__(self, *args):
        self.site = None
        self.builder = None
        self.build_root = None
        self.stack.__exit__(*args)


class MockSite(MockSiteBase):
    """
    Define a mock site for testing
    """

    def __init__(self, files: MockFiles, **kw):
        super().__init__(**kw)
        self.files = files

        # Default settings for testing
        self.settings.SITE_NAME = "Test site"
        self.settings.SITE_URL = "https://www.example.org"
        self.settings.SITE_AUTHOR = "Test User"

    def populate_workdir(self):
        self.settings.PROJECT_ROOT = self.root

        for relpath, content in self.files.items():
            abspath = os.path.join(self.root, relpath)
            os.makedirs(os.path.dirname(abspath), exist_ok=True)
            if isinstance(content, str):
                with open(abspath, "wt") as fd:
                    fd.write(content)
                    os.utime(fd.fileno(), (self.generation_time, self.generation_time))
            elif isinstance(content, bytes):
                with open(abspath, "wb") as fd:
                    fd.write(content)
                    os.utime(fd.fileno(), (self.generation_time, self.generation_time))
            elif isinstance(content, dict):
                with open(abspath, "wt") as fd:
                    fd.write(front_matter.write(content, style="json"))
                    os.utime(fd.fileno(), (self.generation_time, self.generation_time))
            else:
                raise TypeError("content should be a str or bytes")


class ExampleSite(MockSiteBase):
    """
    Site taken from the example/ directory
    """

    def __init__(self, name: str, **kw):
        super().__init__(**kw)
        self.name = name

        # date +%s --date="2020-02-01 16:00 Z"
        self.generation_time = 1580572800

    def populate_workdir(self):
        src = os.path.join(project_root, "example", self.name)
        # dst = os.path.join(self.workdir, "site")
        os.rmdir(self.root)
        shutil.copytree(src, self.root)

        settings_path = os.path.join(self.root, "settings.py")
        if os.path.exists(settings_path):
            self.settings.load(settings_path)
        if self.settings.PROJECT_ROOT is None:
            self.settings.PROJECT_ROOT = self.root
        if self.settings.SITE_URL is None:
            self.settings.SITE_URL = "http://localhost"


class MockSiteTestMixin:
    """
    Test a site built from a mock description
    """

    @contextmanager
    def site(self, mocksite: MockSite | MockFiles, **kw):
        if not isinstance(mocksite, MockSite):
            mocksite = MockSite(mocksite, **kw)
        mocksite.test_case = self
        with mocksite:
            yield mocksite


@contextmanager
def mock_file_stat(overrides: dict[int, Any] | None):
    """
    Override File.stat contents.

    Overrides is a dict like: `{"st_mtime": 12345}`
    """
    if not overrides:
        yield
        return

    real_stat = os.stat
    real_file_from_dir_entry = File.from_dir_entry

    # See https://www.peterbe.com/plog/mocking-os.stat-in-python
    def mock_stat(*args, **kw):
        res = list(real_stat(*args, **kw))
        for k, v in overrides.items():
            res[getattr(stat, k.upper())] = v
        return os.stat_result(res)

    def mock_file_from_dir_entry(dir, entry):
        res = real_file_from_dir_entry(dir, entry)
        st = list(res.stat)
        for k, v in overrides.items():
            st[getattr(stat, k.upper())] = v
        res = File(res.relpath, res.abspath, os.stat_result(st))
        return res

    with mock.patch("staticsite.file.os.stat", new=mock_stat):
        with mock.patch(
            "staticsite.file.File.from_dir_entry", new=mock_file_from_dir_entry
        ):
            yield


def datafile_abspath(relpath):
    test_root = os.path.dirname(__file__)
    return os.path.join(test_root, "data", relpath)


class TracebackHandler(logging.Handler):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.collected = []

    def handle(self, record):
        import traceback

        if record.stack_info is None:
            record.stack_info = traceback.print_stack()
        self.collected.append(record)


@contextmanager
def assert_no_logs(level=logging.WARN):
    handler = TracebackHandler(level=level)
    try:
        root_logger = logging.getLogger()
        root_logger.addHandler(handler)
        yield
    finally:
        root_logger.removeHandler(handler)
    if handler.collected:
        raise AssertionError(f"{len(handler.collected)} unexpected loggings")


class Args:
    """
    Mock argparser namespace initialized with options from constructor
    """

    def __init__(self, **kw):
        self._args = kw

    def __getattr__(self, k):
        return self._args.get(k, None)


class SiteTestMixin:
    """
    Test a real site found on disk
    """

    site_name: str
    site_settings: dict[str, Any] = {}
    site_cls = ExampleSite

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.stack = ExitStack()

        cls.mocksite = cls.stack.enter_context(
            cls.site_cls(name=cls.site_name, settings=cls.site_settings)
        )
        cls.site = cls.mocksite.site

        cls.mocksite.build_site()
        cls.build_root = cls.mocksite.build_root
        cls.build_log = cls.mocksite.builder.build_log

    @classmethod
    def tearDownClass(cls):
        cls.mocksite.__exit__(None, None, None)
        cls.stack.__exit__(None, None, None)
        super().tearDownClass()

    def setUp(self):
        super().setUp()
        self.mocksite.test_case = self

    def tearDown(self):
        self.mocksite.test_case = None
        super().tearDown()

    def page(self, *paths: tuple[str]) -> tuple[Page]:
        """
        Ensure the site has the given page, by path, and return it
        """
        return self.mocksite.page(*paths)

    def assertPagePaths(self, paths: Sequence[str]):
        """
        Check that the list of pages in the site matches the given paths
        """
        return self.mocksite.assertPagePaths(paths)

    def assertBuilt(
        self,
        srcpath: str,
        sitepath: str,
        dstpath: str,
        sample: str | bytes | None = None,
    ):
        """
        Check that page at `srcpath` is in the site at `sitepath` and rendered in `dstpath`.

        Optionally check that the rendered content contains `sample`
        """
        return self.mocksite.assertBuilt(srcpath, sitepath, dstpath, sample)



from __future__ import annotations

import contextlib
import logging
import os
import re
import stat
from collections.abc import Generator
from typing import IO, TYPE_CHECKING, Any, Literal, overload

from .file import File
from .utils import front_matter, open_dir_fd

if TYPE_CHECKING:
    from .site import Site
    from .source_node import SourceNode, SourcePageNode

log = logging.getLogger("fstree")


class Tree:
    """
    Recursive information about a filesystem tree
    """

    def __init__(self, *, site: Site, src: File, node: SourceNode):
        self.site = site
        self.node = node

        # When scanning/itearating, this is set to a path fd opened to this
        # directory
        self.dir_fd: int | None = None

        # The directory corresponding to this node
        self.src = src

        # Files in this directory
        self.files: dict[str, File] = {}

        # Subdirectories
        self.sub: dict[str, Tree] = {}

        # Rules for ignoring files
        self.ignore_rules: list[re.Pattern[str]] = []

    def print(self, lead: str = "", file: IO[str] | None = None) -> None:
        print(f"{lead}", file=file)
        for name, src in self.files.items():
            print(f"{lead}→ {name}", file=file)
        for name, tree in self.sub.items():
            print(f"{lead}📂 {name}", file=file)
            tree.print(lead + "  ", file=file)

    @contextlib.contextmanager
    def open_tree(self) -> Generator[None, None, None]:
        """
        Open self.dir_fd for the duration of this context manager
        """
        with open_dir_fd(self.src.abspath) as dir_fd:
            self.dir_fd = dir_fd
            try:
                yield
            finally:
                self.dir_fd = None

    @contextlib.contextmanager
    def open_subtree(self, name: str, tree: Tree) -> Generator[None, None, None]:
        """
        Open tree.dir_fd for the duration of this context manager
        """
        with open_dir_fd(name, self.dir_fd) as subdir_fd:
            tree.dir_fd = subdir_fd
            try:
                yield
            finally:
                tree.dir_fd = None

    def _scandir(self) -> None:
        """
        Scan the contents of a directory, filling in structures but without
        recursing
        """
        raise NotImplementedError(f"{self.__class__.__name__}._scandir not implemented")

    def _apply_ignore_rules(self) -> None:
        """
        Remove from self.files all entries that match self.ignore_rules
        """
        for name in list(self.files.keys()):
            for rule in self.ignore_rules:
                if rule.match(name):
                    del self.files[name]

    def scan(self) -> None:
        """
        Scan directory contents, recursively
        """
        self._scandir()
        # Recurse into subdirectories
        for name, tree in self.sub.items():
            with self.open_subtree(name, tree):
                tree.scan()

    def populate_node(self) -> None:
        """
        Recursively popuplate the node with information from this tree
        """
        raise NotImplementedError(
            f"{self.__class__.__name__}.populate_node not implemented"
        )

    @overload
    def open(self, name: str, mode: Literal["rt"]) -> IO[str]:
        ...

    @overload
    def open(self, name: str, mode: Literal["rb"]) -> IO[bytes]:
        ...

    def open(self, name: str, mode: str) -> IO[str] | IO[bytes]:
        """
        Open a file contained in this directory
        """
        if self.dir_fd is None:
            raise RuntimeError("Tree.open called without a dir_fd")

        def _file_opener(fname: str, flags: int) -> int:
            return os.open(fname, flags, dir_fd=self.dir_fd)

        return open(name, mode=mode, opener=_file_opener)


class PageTree(Tree):
    """
    Filesystem tree that can contain pages and assets
    """

    def __init__(self, *, site: Site, src: File, node: SourcePageNode):
        super().__init__(site=site, src=src, node=node)
        self.node: SourcePageNode

        # Rules for assigning metadata to subdirectories
        self.dir_rules: list[tuple[re.Pattern[str], dict[str, Any]]] = []

        # Rules for assigning metadata to files
        self.file_rules: list[tuple[re.Pattern[str], dict[str, Any]]] = []

    def _take_dir_rules(self, meta: dict[str, Any]) -> None:
        """
        Remove file and directory rules from the meta dict, and compile them
        for this Tree
        """
        from .page_filter import compile_page_match

        # Compile directory matching rules
        dir_meta = meta.pop("dirs", None)
        if dir_meta is None:
            dir_meta = {}
        self.dir_rules.extend((compile_page_match(k), v) for k, v in dir_meta.items())

        # Compute file matching rules
        file_meta = meta.pop("files", None)
        if file_meta is None:
            file_meta = {}
        self.file_rules.extend((compile_page_match(k), v) for k, v in file_meta.items())

        # Compute file ignore rules
        ignore = meta.pop("ignore", None)
        if ignore is None:
            ignore = []
        self.ignore_rules.extend(compile_page_match(k) for k in ignore)

    def _load_dir_meta(self) -> dict[str, Any]:
        """
        Let features add to directory metadata

        This for example lets metadata of an index.md do the same work as a
        .staticfile file
        """
        res = {}
        for feature in self.site.features.ordered():
            meta = feature.load_dir_meta(self)
            if meta is not None:
                self._take_dir_rules(meta)
                res.update(meta)
        # Filter out metadata entries that were specific to the index page of the directory
        return {k: v for k, v in res.items() if k in self.node._fields}

    def _scandir(self) -> None:
        subdirs: list[tuple[str, File]] = []
        with os.scandir(self.dir_fd) as entries:
            for entry in entries:
                # Note: is_dir, is_file, and stat, follow symlinks by default
                if entry.is_dir():
                    if entry.name.startswith("."):
                        # Skip hidden directories
                        continue
                    # Take note of directories
                    subdirs.append((entry.name, File.from_dir_entry(self.src, entry)))
                elif entry.name == ".staticsite":
                    # Load dir metadata from .staticsite
                    with self.open(entry.name, "rt") as fd:
                        fmt, meta = front_matter.read_whole(fd)
                        # Honor skip: yes, completely skipping this subdir
                        if meta.get("skip", False):
                            self.files.clear()
                            return
                        self._take_dir_rules(meta)
                        self.node.update_fields(meta)
                elif entry.name.startswith("."):
                    # Skip hidden files
                    continue
                else:
                    # Take note of files
                    try:
                        self.files[entry.name] = File.from_dir_entry(self.src, entry)
                    except FileNotFoundError:
                        log.warning(
                            "%s: cannot stat() file: broken symlink?",
                            os.path.join(self.src.abspath, entry.name),
                        )

        # Let features add to directory metadata
        self.node.update_fields(self._load_dir_meta())

        # Apply ignore rules
        self._apply_ignore_rules()

        # Instantiate subtrees
        for name, src in subdirs:
            meta = {}

            # Compute metadata for this directory
            for pattern, dmeta in self.dir_rules:
                if pattern.match(name):
                    meta.update(dmeta)

            node: SourceNode
            tree: Tree
            if meta.get("asset"):
                node = self.node.asset_child(name, src)
                tree = AssetTree(site=self.site, src=src, node=node)
            else:
                node = self.node.page_child(name, src)
                tree = PageTree(site=self.site, src=src, node=node)

            # Inherit ignore rules
            tree.ignore_rules.extend(self.ignore_rules)

            node.update_fields(meta)
            self.sub[name] = tree

    def populate_node(self) -> None:
        # Add the metadata scanned for this directory

        # Compute metadata for files
        files_meta: dict[str, tuple[dict[str, Any], File]] = {}
        for fname, src in self.files.items():
            res: dict[str, Any] = {}
            for pattern, meta in self.file_rules:
                if pattern.match(fname):
                    res.update(meta)

            # Handle assets right away
            if res.get("asset"):
                # print(f"PageTree.populate_node  add asset {fname}")
                self.node.add_asset(src=src, name=fname)
            else:
                # print(f"PageTree.populate_node  enqueue file {fname}")
                files_meta[fname] = (res, src)

        # Recurse into subtrees
        for name, tree in self.sub.items():
            with self.open_subtree(name, tree):
                tree.populate_node()

        # Let features pick their files
        # print(f"PageTree.populate_node  initial files to pick {files_meta.keys()}")
        for handler in self.site.features.ordered():
            handler.load_dir(self.node, self, files_meta)
            if not files_meta:
                break
        # print(f"PageTree.populate_node  remaining files to pick {files_meta.keys()}")

        # Use everything else as an asset
        # TODO: move into an asset feature?
        for fname, (file_meta, src) in files_meta.items():
            if src.stat and stat.S_ISREG(src.stat.st_mode):
                log.debug("Loading static file %s", src.relpath)
                self.node.add_asset(src=src, name=fname)

        self.node.prune_empty_subnodes()


class RootPageTree(PageTree):
    """
    Special behaviour of the root of the directory hierarchy
    """

    def _load_dir_meta(self) -> dict[str, Any]:
        res = super()._load_dir_meta()
        if (
            (title := res.get("title"))
            and "site_name" not in res
            and not self.node.site_name
        ):
            res["site_name"] = title
        return res


class AssetTree(Tree):
    """
    Filesystem tree that only contains assets
    """

    def __init__(self, *, site: Site, src: File, node: SourceNode):
        super().__init__(site=site, src=src, node=node)

    def _scandir(self) -> None:
        with os.scandir(self.dir_fd) as entries:
            for entry in entries:
                if entry.name.startswith("."):
                    # Skip hidden directories
                    continue

                # Note: is_dir, is_file, and stat, follow symlinks by default
                if entry.is_dir():
                    # Take note of directories
                    src = File.from_dir_entry(self.src, entry)
                    tree = AssetTree(
                        site=self.site,
                        src=src,
                        node=self.node.asset_child(entry.name, src),
                    )
                    tree.ignore_rules.extend(self.ignore_rules)
                    self.sub[entry.name] = tree
                else:
                    # Take note of files
                    try:
                        self.files[entry.name] = File.from_dir_entry(self.src, entry)
                    except FileNotFoundError:
                        log.warning(
                            "%s: cannot stat() file: broken symlink?",
                            os.path.join(self.src.abspath, entry.name),
                        )

        # Apply ignore rules
        self._apply_ignore_rules()

    def populate_node(self) -> None:
        # Recurse into subdirectories
        for name, tree in self.sub.items():
            # Recursively descend into the directory
            with self.open_subtree(name, tree):
                tree.populate_node()

        # Load every file as an asset
        for fname, src in self.files.items():
            if stat.S_ISREG(src.stat.st_mode):
                log.debug("Loading static file %s", src.relpath)
                self.node.add_asset(src=src, name=fname)

        self.node.prune_empty_subnodes()



import os
from unittest import TestCase

from . import utils as test_utils


class TestNav(test_utils.MockSiteTestMixin, TestCase):
    """
    Test metadata collected on site load
    """

    def test_dir(self):
        self.maxDiff = None

        files = {
            "index.md": {"nav": ["about.md"]},
            "about.md": {},
            "dir1/dir2/dir3/page.md": {},
        }

        with self.site(files) as mocksite:
            page = mocksite.page("dir1/dir2/dir3/page")
            self.assertEqual(
                page.to_dict(),
                {
                    "src": {
                        "relpath": "dir1/dir2/dir3/page.md",
                        "abspath": os.path.join(
                            mocksite.site.content_root, "dir1/dir2/dir3/page.md"
                        ),
                    },
                    "build_path": "dir1/dir2/dir3/page/index.html",
                    "site_path": "dir1/dir2/dir3/page",
                    "meta": {
                        "author": "Test User",
                        "copyright": "© 2019 Test User",
                        "date": "2019-06-01 12:30:00+02:00",
                        "draft": False,
                        "indexed": True,
                        "nav": ["MarkdownPage(about)"],
                        "syndicated": True,
                        "syndication_date": "2019-06-01 12:30:00+02:00",
                        "site_name": "Test site",
                        "site_url": "https://www.example.org",
                        "template": "page.html",
                        "template_copyright": "compiled:None",
                        "title": "Test site",
                        "related": {},
                    },
                    "type": "markdown",
                },
            )



from __future__ import annotations

from unittest import TestCase

from . import utils as test_utils


class TestPage(test_utils.MockSiteTestMixin, TestCase):
    def test_resolve_path(self):
        files = {
            "toplevel.md": {},
            "lev1/page1.md": {},
            "lev1/page2.md": {},
            "lev1/lev2/page1.md": {},
        }
        with self.site(files) as mocksite:
            index, toplevel, lev1page1, lev1page2, lev2page1 = mocksite.page(
                "", "toplevel", "lev1/page1", "lev1/page2", "lev1/lev2/page1"
            )

            self.assertEqual(lev1page1.resolve_path("/lev1/page2.md"), lev1page2)
            self.assertEqual(lev1page1.resolve_path("/lev1/page2"), lev1page2)
            self.assertEqual(lev1page1.resolve_path("page2.md"), lev1page2)
            self.assertEqual(lev1page1.resolve_path("page2"), lev1page2)

            self.assertEqual(lev1page1.resolve_path("../toplevel.md"), toplevel)
            self.assertEqual(lev1page1.resolve_path("../toplevel"), toplevel)
            self.assertEqual(lev1page1.resolve_path("/toplevel.md"), toplevel)
            self.assertEqual(lev1page1.resolve_path("/toplevel"), toplevel)

            self.assertEqual(lev1page1.resolve_path("/lev1/lev2/page1.md"), lev2page1)
            self.assertEqual(lev1page1.resolve_path("/lev1/lev2/page1"), lev2page1)
            self.assertEqual(lev1page1.resolve_path("lev2/page1.md"), lev2page1)
            self.assertEqual(lev1page1.resolve_path("lev2/page1"), lev2page1)

            self.assertEqual(lev1page1.resolve_path("/"), index)

    def test_meta(self):
        files = {
            "index.md": {"title": "test"},
        }
        with self.site(files) as mocksite:
            index = mocksite.page("")
            # Direct field access
            self.assertEqual(index.title, "test")
            self.assertEqual(index.front_matter, {"title": "test"})

            # Dict-like field access via page.meta
            self.assertIn("title", index.meta)
            self.assertNotIn("test", index.meta)
            # Internal fields are hidden
            self.assertNotIn("front_matter", index.meta)

            self.assertEqual(index.meta["title"], "test")
            self.assertEqual(index.meta.get("title"), "test")

            with self.assertRaises(KeyError):
                index.meta["test"]
            self.assertIsNone(index.meta.get("test"))

            with self.assertRaises(KeyError):
                index.meta["front_matter"]
            self.assertIsNone(index.meta.get("front_matter"))

    def test_iter_pages(self):
        files = {
            "index.md": {},
            "index.txt": "",
            "tags.taxonomy": {},
            "lev1/page1.md": {"tags": ["a"]},
        }
        with self.site(files) as mocksite:
            (
                index,
                asset,
                tags,
                tag_a,
                lev1,
                lev1page1,
            ) = mocksite.page("", "index.txt", "tags", "tags/a", "lev1", "lev1/page1")
            rss, atom, archive = mocksite.page(
                "tags/a/index.rss", "tags/a/index.atom", "tags/a/archive"
            )

            # Full iteration
            assets = []
            pages = []
            for page in mocksite.site.iter_pages():
                if page.TYPE == "asset":
                    assets.append(page)
                else:
                    pages.append(page)
            self.assertGreaterEqual(len(assets), 10)
            self.assertCountEqual(
                pages, [index, tags, tag_a, lev1, lev1page1, rss, atom, archive]
            )

            # Skip static
            assets = []
            pages = []
            for page in mocksite.site.iter_pages(static=False):
                if page.TYPE == "asset":
                    assets.append(page)
                else:
                    pages.append(page)
            self.assertCountEqual(assets, [asset])
            self.assertCountEqual(
                pages, [index, tags, tag_a, lev1, lev1page1, rss, atom, archive]
            )

            # Only source pages
            assets = []
            pages = []
            for page in mocksite.site.iter_pages(static=False, source_only=True):
                if page.TYPE == "asset":
                    assets.append(page)
                else:
                    pages.append(page)
            self.assertCountEqual(assets, [asset])
            self.assertCountEqual(pages, [index, tags, lev1page1])



from __future__ import annotations

import contextlib
import logging
from collections.abc import Generator
from typing import TYPE_CHECKING, Any, NamedTuple
from urllib.parse import urlparse, urlunparse

from .page import PageNotFoundError, SourcePage

if TYPE_CHECKING:
    import urllib.parse

    from .page import Page

log = logging.getLogger("markdown")


class ResolvedLink(NamedTuple):
    page: Page
    site_path: str


class ResolvedPage(NamedTuple):
    page: Page
    url: urllib.parse.ParseResult


class LinkResolver:
    """
    Caching backend for resolving internal URLs in rendered content
    """

    def __init__(self) -> None:
        self.page: Page | None = None
        self.absolute: bool = False
        self.substituted: dict[str, ResolvedLink] = {}
        self.external_links: set[str] = set()

    def set_page(self, page: Page, absolute: bool = False) -> None:
        self.page = page
        self.absolute = absolute
        self.substituted = {}
        self.external_links = set()

    def load_cache(self, paths: list[tuple[str, str]]) -> bool:
        # If the destination of links has changed, drop the cached version.
        # This will as a side effect prime the link resolver cache,
        # avoiding links from being looked up again during rendering
        for src, dest in paths:
            if (resolved := self.resolve_page(src)) is None:
                return False
            if resolved.page.site_path != dest:
                return False
        return True

    def to_cache(self) -> list[tuple[str, str]]:
        return [(k, v.site_path) for k, v in self.substituted.items()]

    def resolve_page(self, url: str) -> ResolvedPage | None:
        if self.page is None:
            raise RuntimeError(
                "LinkResolver.resolve_page called before LinkResolver.set_page"
            )
        parsed = urlparse(url)

        # If it's an absolute url, leave it unchanged
        if parsed.scheme or parsed.netloc:
            self.external_links.add(url)
            return None

        # If it's an anchor inside the page, leave it unchanged
        if not parsed.path:
            return None

        # Try with cache
        if (resolved := self.substituted.get(parsed.path)) is not None:
            return ResolvedPage(resolved.page, parsed)

        # Resolve as a path
        try:
            page = self.page.resolve_path(parsed.path)
        except PageNotFoundError as e:
            log.warning("%s: %s", self.page, e)
            return None

        # Cache the page site_path
        self.substituted[url] = ResolvedLink(page, page.site_path)

        return ResolvedPage(page, parsed)

    def resolve_url(self, url: str) -> str:
        """
        Resolve internal URLs.

        Returns None if the URL does not need changing, else returns the new URL.
        """
        if self.page is None:
            raise RuntimeError(
                "LinkResolver.resolve_page called before LinkResolver.set_page"
            )
        if (resolved := self.resolve_page(url)) is None:
            return url

        new_url = self.page.url_for(resolved.page, absolute=self.absolute)
        dest = urlparse(new_url)

        return urlunparse(
            (
                dest.scheme,
                dest.netloc,
                dest.path,
                resolved.url.params,
                resolved.url.query,
                resolved.url.fragment,
            )
        )


class MarkupFeature:
    def __init__(self, *args: Any, **kw: Any):
        super().__init__(*args, **kw)
        self.link_resolver = LinkResolver()


class MarkupRenderContext:
    """
    State held during rendering of a page markup
    """

    def __init__(self, page: MarkupPage, cache_key: str):
        self.page = page
        self.link_resolver = page.feature.link_resolver
        self.cache_key = cache_key
        self.cache: dict[str, Any]

    def load(self) -> None:
        if (cache := self.page.feature.render_cache.get(self.cache_key)) is None:
            self.reset_cache()
            return

        if cache["mtime"] != self.page.src.stat.st_mtime:
            # If the source has changed, drop the cached version
            self.reset_cache()
            return

        if (paths := cache.get("paths")) is None or not self.link_resolver.load_cache(
            paths
        ):
            self.reset_cache()
            return

        self.cache = cache

    def reset_cache(self) -> None:
        self.cache = {
            "mtime": self.page.src.stat.st_mtime,
        }

    def save(self) -> None:
        self.cache["paths"] = self.link_resolver.to_cache()
        self.page.feature.render_cache.put(self.cache_key, self.cache)


class MarkupPage(SourcePage):
    """
    Page which renders content from text markup.

    This is a base for pages like Markdown or Rst
    """

    def __init__(self, *, feature: MarkupFeature, **kw: Any):
        super().__init__(**kw)
        self.feature = feature

    @contextlib.contextmanager
    def markup_render_context(
        self, cache_key: str, absolute: bool = False
    ) -> Generator[MarkupRenderContext, None, None]:
        self.feature.link_resolver.set_page(self, absolute)
        render_context = MarkupRenderContext(self, cache_key)
        render_context.load()
        yield render_context
        render_context.save()



from __future__ import annotations

import datetime
from unittest import TestCase

from . import utils as test_utils


def dt(*args):
    return str(datetime.datetime(*args))


class TestSeries(test_utils.MockSiteTestMixin, TestCase):
    def test_site(self):
        files = {
            "series.taxonomy": "",
            "seriesa1.md": {
                "series": "seriesa",
                "title": "Series A",
                "date": dt(2016, 1, 1),
            },
            "seriesa2.md": {"date": dt(2016, 1, 2), "series": "seriesa", "title": "A2"},
            "seriesa3.md": {
                "date": dt(2016, 1, 3),
                "series": "seriesa",
                "title": "A3",
                "series_title": "Series A part Two",
            },
            "seriesa4.md": {"date": dt(2016, 1, 4), "series": "seriesa", "title": "A4"},
            "seriesb1.md": {
                "date": dt(2016, 1, 1),
                "series": "seriesb",
                "title": "Series B Intro",
                "series_title": "Series B",
            },
            "seriesb2.md": {"date": dt(2016, 1, 2), "series": "seriesb", "title": "B2"},
            "seriesc1.md": {
                "date": dt(2016, 1, 1),
                "title": "Series C",
                "series": "seriesc",
            },
            "noseries.md": {"date": dt(2016, 1, 1), "title": "Other things"},
        }

        with self.site(files) as mocksite:
            seriesa1 = mocksite.page("seriesa1")
            seriesa2 = mocksite.page("seriesa2")
            seriesa3 = mocksite.page("seriesa3")
            seriesa4 = mocksite.page("seriesa4")
            seriesb1 = mocksite.page("seriesb1")
            seriesb2 = mocksite.page("seriesb2")
            seriesc1 = mocksite.page("seriesc1")
            noseries = mocksite.page("noseries")
            series = mocksite.page("series")

            # Check that series have been built (check in tags)
            self.assertIn("seriesa", series.categories)
            self.assertIn("seriesb", series.categories)
            self.assertIn("seriesc", series.categories)
            self.assertEqual(len(series.categories), 3)

            # Check the contents of series (check in tags)
            self.assertEqual(
                series.categories["seriesa"].pages,
                [seriesa1, seriesa2, seriesa3, seriesa4],
            )
            self.assertEqual(series.categories["seriesb"].pages, [seriesb1, seriesb2])
            self.assertEqual(series.categories["seriesc"].pages, [seriesc1])

            # Check computed series metadata
            self.assertEqual(
                series.categories["seriesa"].series_info,
                {
                    "pages": [seriesa1, seriesa2, seriesa3, seriesa4],
                    "length": 4,
                    "first": seriesa1,
                    "last": seriesa4,
                    "title": "Series A",
                },
            )

            s = series.categories["seriesa"].sequence(seriesa1)
            self.assertEqual(s["title"], "Series A")
            self.assertEqual(s["prev"], None)
            self.assertEqual(s["next"], seriesa2)
            self.assertEqual(s["first"], seriesa1)
            self.assertEqual(s["last"], seriesa4)
            self.assertEqual(s["index"], 1)
            self.assertEqual(s["length"], 4)

            s = series.categories["seriesa"].sequence(seriesa2)
            self.assertEqual(s["title"], "Series A")
            self.assertEqual(s["prev"], seriesa1)
            self.assertEqual(s["next"], seriesa3)
            self.assertEqual(s["first"], seriesa1)
            self.assertEqual(s["last"], seriesa4)
            self.assertEqual(s["index"], 2)
            self.assertEqual(s["length"], 4)

            s = series.categories["seriesa"].sequence(seriesa3)
            self.assertEqual(s["title"], "Series A part Two")
            self.assertEqual(s["prev"], seriesa2)
            self.assertEqual(s["next"], seriesa4)
            self.assertEqual(s["first"], seriesa1)
            self.assertEqual(s["last"], seriesa4)
            self.assertEqual(s["index"], 3)
            self.assertEqual(s["length"], 4)

            s = series.categories["seriesa"].sequence(seriesa4)
            self.assertEqual(s["title"], "Series A part Two")
            self.assertEqual(s["prev"], seriesa3)
            self.assertEqual(s["next"], None)
            self.assertEqual(s["first"], seriesa1)
            self.assertEqual(s["last"], seriesa4)
            self.assertEqual(s["index"], 4)
            self.assertEqual(s["length"], 4)

            self.assertEqual(
                series.categories["seriesb"].series_info,
                {
                    "pages": [seriesb1, seriesb2],
                    "length": 2,
                    "first": seriesb1,
                    "last": seriesb2,
                    "title": "Series B",
                },
            )

            s = series.categories["seriesb"].sequence(seriesb1)
            self.assertEqual(s["title"], "Series B")
            self.assertEqual(s["prev"], None)
            self.assertEqual(s["next"], seriesb2)
            self.assertEqual(s["first"], seriesb1)
            self.assertEqual(s["last"], seriesb2)
            self.assertEqual(s["index"], 1)
            self.assertEqual(s["length"], 2)

            s = series.categories["seriesb"].sequence(seriesb2)
            self.assertEqual(s["title"], "Series B")
            self.assertEqual(s["prev"], seriesb1)
            self.assertEqual(s["next"], None)
            self.assertEqual(s["first"], seriesb1)
            self.assertEqual(s["last"], seriesb2)
            self.assertEqual(s["index"], 2)
            self.assertEqual(s["length"], 2)

            s = series.categories["seriesc"].sequence(seriesc1)
            self.assertEqual(s["title"], "Series C")
            self.assertEqual(s["prev"], None)
            self.assertEqual(s["next"], None)
            self.assertEqual(s["first"], seriesc1)
            self.assertEqual(s["last"], seriesc1)
            self.assertEqual(s["index"], 1)
            self.assertEqual(s["length"], 1)

            self.assertIsNone(series.categories["seriesa"].sequence(noseries))
            self.assertIsNone(series.categories["seriesb"].sequence(noseries))
            self.assertIsNone(series.categories["seriesc"].sequence(noseries))



from __future__ import annotations

import itertools
import logging
import mimetypes
import os
from collections.abc import Iterable, Sequence
from typing import TYPE_CHECKING, Any, Union, cast

from staticsite import fields
from staticsite.feature import Feature, PageTrackingMixin, TrackedField
from staticsite.page import AutoPage, ChangeExtent, ImagePage, Page, SourcePage
from staticsite.render import RenderedElement, RenderedFile
from staticsite.utils.images import ImageScanner

if TYPE_CHECKING:
    from staticsite import file, fstree
    from staticsite.source_node import SourcePageNode

log = logging.getLogger("images")


def basename_no_ext(pathname: str) -> str:
    """
    Return the basename of pathname, without extension
    """
    return os.path.splitext(os.path.basename(pathname))[0]


class ImageField(TrackedField[Page, Union[str, "Image"]]):
    """
    Image used for this post.

    It is set to a path to an image file relative to the current page.

    During the crossreference phase, it is resolved to the corresponding
    [image page](images.md).

    If not set, and an image exists with the same name as the page (besides the
    extension), that image is used.
    """

    tracked_by = "images"

    def _clean(self, page: Page, value: Any) -> str | Image:
        if isinstance(value, str):
            return value
        elif isinstance(value, Image):
            return value
        else:
            raise TypeError(
                f"invalid value of type {type(value)} for {page!r}.{self.name}:"
                " expecting str or Image"
            )


class ImagePageMixin(Page):
    image = ImageField()


class Images(PageTrackingMixin[ImagePageMixin], Feature):
    """
    Handle images in content directory.

    Files with an `image/*` mime type found in content directories are scanned for
    metadata and used as image pages.

    Their behaviour is similar to the one of static assets, but templates and other
    staticsite facilities can make use of the metadata resolved.


    ## Metadata extracted from images

    These are the current metadata elements filled from image data, when available:

    * `page.meta.width`: image width in pixels
    * `page.meta.height`: image height in pixels
    * `page.meta.lat`: latitude from EXIF tags, as a signed floating point number
      of degrees
    * `page.meta.lon`: longitude from EXIF tags, as a signed floating point number
      of degrees
    * `page.meta.title`: content of the `ImageDescription` EXIF tag
    * `page.meta.author`: content of the `Artist` EXIF tag
    * `page.meta.image_orientation`: content of the
      [`ImageOrientation` EXIF tag](https://www.impulseadventure.com/photo/exif-orientation.html)


    ## Image associated to another page

    If you have an image sharing the file name with another page (for example, an
    image called `example.jpg` next to a page called `example.md`), then the page
    automatically gets a `page.meta.image` value pointing to the image.

    You can use this, for example, to easily provide an image for blog posts.


    ## Scaled versions of images

    [Theme](../theme.md) configurations can specify a set of scaled versions of images
    to generate:

    ```yaml
    # Scaled down widths of images that will be generated
    image_sizes:
      medium:
        width: 600
      small:
        width: 480
      thumbnail:
        width: 128
    ```

    Given a `image.jpg` image file, this will generate `image-medium.jpg`,
    `image-small.jpg` and `image-thumbnail.jpg`.

    Using this feature, the [`img_for` template function](templates.md) will
    automatically generate `srcset` attributes to let the browser choose which
    version to load.

    Scaled versions of the image will be available in the image's [`related`
    field](../fields/related.md).
    """

    def __init__(self, *args: Any, **kw: Any) -> None:
        super().__init__(*args, **kw)
        mimetypes.init()
        self.scanner = ImageScanner(self.site.caches.get("images_meta"))
        # Nodes that contain images
        self.images: set[Image] = set()

    def get_page_bases(self, page_cls: type[Page]) -> Sequence[type[Page]]:
        return (ImagePageMixin,)

    def get_used_page_types(self) -> list[type[Page]]:
        return [Image, ScaledImage]

    def load_dir(
        self,
        node: SourcePageNode,
        directory: fstree.Tree,
        files: dict[str, tuple[dict[str, Any], file.File]],
    ) -> list[Page]:
        taken: list[str] = []
        pages: list[Page] = []
        for fname, (kwargs, src) in files.items():
            base, ext = os.path.splitext(fname)
            mimetype = mimetypes.types_map.get(ext)
            if mimetype is None:
                continue

            if not mimetype.startswith("image/"):
                continue

            taken.append(fname)

            img_meta = self.scanner.scan(src, mimetype)
            kwargs.update(img_meta)

            page = node.create_source_page_as_file(
                page_cls=Image, src=src, mimetype=mimetype, dst=fname, **kwargs
            )

            if page is not None:
                pages.append(page)
                self.images.add(page)

        for fname in taken:
            del files[fname]

        return pages

    def generate(self) -> None:
        for image in self.images:
            # Look at theme's image_sizes and generate ScaledImage pages
            image_sizes = self.site.theme.meta.get("image_sizes")
            if image_sizes:
                for name, info in image_sizes.items():
                    width = image.width
                    if width is None:
                        # SVG images, for example, don't have width
                        continue
                    if info["width"] >= width:
                        continue
                    rel_kwargs = dict(info)
                    rel_kwargs["related"] = {}

                    base, ext = os.path.splitext(os.path.basename(image.src.relpath))
                    scaled_fname = f"{base}-{name}{ext}"

                    image.node.create_auto_page_as_file(
                        page_cls=ScaledImage,
                        created_from=image,
                        mimetype=image.mimetype,
                        name=name,
                        info=info,
                        dst=scaled_fname,
                        **rel_kwargs,
                    )

    def crossreference(self) -> None:
        # Resolve image from strings to Image pages
        for page in self.tracked_pages:
            val = page.image
            if isinstance(val, str):
                resolved = page.resolve_path(val)
                if not isinstance(resolved, (Image, ScaledImage)):
                    log.warning(
                        "%s: image field resolves to %s which is not an image page",
                        self,
                        page.image,
                    )
                else:
                    page.image = resolved

        # If an image exists with the same basename as a page, auto-add an
        # "image" metadata to it
        for image in self.images:
            name = basename_no_ext(image.src.relpath)
            # print(f"Images.analyze {image=!r} {image.node.name=!r} {image.node.page=!r}")
            pages: Iterable[ImagePageMixin]
            pages = cast(Iterable[ImagePageMixin], image.node.build_pages.values())
            if image.node.sub is not None:
                pages = itertools.chain(
                    pages,
                    (
                        cast(ImagePageMixin, subnode.page)
                        for subnode in image.node.sub.values()
                        if subnode.page
                    ),
                )

            # Find pages matching this image's name
            for page in pages:
                # print(f"Images.analyze  check {page=!r} {page.src=!r}")
                if not isinstance(page, SourcePage):
                    # Don't associate to generated pages
                    continue
                if page.src.relpath == image.src.relpath:
                    # Don't associate to variants of this image
                    continue
                if basename_no_ext(page.src.relpath) == name:
                    # Don't add if already set
                    if not page.image and basename_no_ext(page.src.relpath) == name:
                        # print(f"Images.analyze  add {image!r}")
                        page.image = image
                    break


class Image(ImagePage, SourcePage):
    """
    An image as found in the source directory
    """

    TYPE = "image"

    lat = fields.Float[Page](doc="Image latitude")
    lon = fields.Float[Page](doc="Image longitude")
    image_orientation = fields.Int[Page](doc="Image orientation")

    def __init__(self, *args: Any, mimetype: str, **kw: Any):
        super().__init__(*args, **kw)
        self.mimetype = mimetype
        # self.date = self.site.localized_timestamp(self.src.stat.st_mtime)

    def render(self, **kw: Any) -> RenderedElement:
        return RenderedFile(self.src)


class RenderedScaledImage(RenderedElement):
    def __init__(self, src: file.File, width: int, height: int):
        self.src = src
        self.width = width
        self.height = height

    def write(self, *, name: str, dir_fd: int, old: os.stat_result | None) -> None:
        # If target exists and mtime is ok, keep it
        if old and old.st_mtime >= self.src.stat.st_mtime:
            return
        import PIL

        with PIL.Image.open(self.src.abspath) as img:
            img = img.resize((self.width, self.height))
            with self.dirfd_open(name, "wb", dir_fd=dir_fd) as out:
                img.save(out)

    def content(self) -> bytes:
        with open(self.src.abspath, "rb") as fd:
            return fd.read()


class ScaledImage(ImagePage, AutoPage):
    """
    Scaled version of an image
    """

    TYPE = "scaledimage"

    # Note: from Python 3.12, we can parameterize Page on the type of
    # created_from, defaulting to Page: https://peps.python.org/pep-0696/
    #
    # Now if we did that we'd have to specify a third type explicitly every
    # time we describe a page, and it's easier to work around it with cast.
    # With 3.12 we can remove the casts in favour of an extra type in the page
    # definition

    def __init__(
        self, *args: Any, mimetype: str, name: str, info: dict[str, Any], **kw: Any
    ):
        super().__init__(*args, **kw)
        self.name = name
        created_from = cast(Image, self.created_from)
        self.date = created_from.date
        if title := created_from.title:
            self.title = title

        if "height" not in kw:
            self.height = round(
                created_from.height * (info["width"] / created_from.width)
            )

        created_from.add_related(name, self)

    def render(self, **kw: Any) -> RenderedElement:
        return RenderedScaledImage(
            cast(Image, self.created_from).src, self.width, self.height
        )

    def _compute_change_extent(self) -> ChangeExtent:
        if self.created_from is None:
            raise AssertionError(f"ScaledImage {self!r} has empty created_from")
        return self.created_from.change_extent


FEATURES = {
    "images": Images,
}



from __future__ import annotations

import logging
import os
from collections.abc import Generator
from typing import NamedTuple

log = logging.getLogger("contents")


class File(NamedTuple):
    """
    Information about a file in the file system.

    If stat is not None, the file exists and these are its stats.
    """

    # Relative path to root
    relpath: str
    # Absolute path to the file
    abspath: str
    # File stats if the file exists, else None
    stat: os.stat_result

    def __str__(self) -> str:
        if self.abspath:
            return self.abspath
        else:
            return self.relpath

    @classmethod
    def from_dir_entry(cls, dir: File, entry: os.DirEntry[str]) -> File:
        return cls(
            relpath=os.path.join(dir.relpath, entry.name),
            abspath=os.path.join(dir.abspath, entry.name),
            stat=entry.stat(),
        )

    @classmethod
    def with_stat(cls, relpath: str, abspath: str) -> File:
        return cls(relpath, abspath, os.stat(abspath))

    @classmethod
    def scan(
        cls, abspath: str, follow_symlinks: bool = False, ignore_hidden: bool = False
    ) -> Generator[File, None, None]:
        """
        Scan tree_root, generating relative paths based on it
        """
        for root, dnames, fnames, dirfd in os.fwalk(
            abspath, follow_symlinks=follow_symlinks
        ):
            if ignore_hidden:
                # Ignore hidden directories
                filtered = [d for d in dnames if not d.startswith(".")]
                if len(filtered) != len(dnames):
                    dnames[::] = filtered

            for f in fnames:
                # Ignore hidden files
                if ignore_hidden and f.startswith("."):
                    continue

                try:
                    st = os.stat(f, dir_fd=dirfd)
                except FileNotFoundError:
                    # Skip broken links
                    continue
                relpath = os.path.join(root, f)
                yield cls(
                    relpath=relpath, abspath=os.path.join(abspath, relpath), stat=st
                )



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from staticsite.page import AutoPage, ChangeExtent, SourcePage, TemplatePage

from .data import LinkCollection

if TYPE_CHECKING:
    from staticsite.node import Node

    from . import Links

log = logging.getLogger("links")


class LinkIndexPage(TemplatePage, SourcePage):
    """
    Root page for the browseable archive of annotated external links in the
    site
    """

    TYPE = "links_index"
    TEMPLATE = "data-links.html"

    def __init__(
        self,
        *args: Any,
        node: Node,
        links: Links,
        link_collection: LinkCollection,
        **kw: Any,
    ):
        kw.setdefault("nav_title", node.name.capitalize())
        kw.setdefault("title", "All links shared in the site")
        super().__init__(*args, node=node, **kw)
        # Reference to the Feature with the aggregated link collection
        self.feature_links = links
        self.link_collection = link_collection

        self.by_tag: dict[str, LinksTagPage] = {}

    #    def to_dict(self):
    #        from staticsite.utils import dump_meta
    #        res = super().to_dict()
    #        res["name"] = self.name
    #        res["categories"] = dump_meta(self.categories)
    #        res["category_meta"] = dump_meta(self.category_meta)
    #        return res

    def organize(self) -> None:
        pages = []
        for tag, links in self.feature_links.by_tag.items():
            name = tag + "-links"

            if name in self.node.sub:
                # A page already exists
                continue

            page = self.node.create_auto_page_as_path(
                created_from=self,
                name=name,
                page_cls=LinksTagPage,
                data_type="links",
                title=f"{tag} links",
                link_collection=links,
            )
            if page is not None:
                self.by_tag[tag] = page
                pages.append(page)

        # Set self.meta.pages to the sorted list of categories
        pages.sort(key=lambda x: x.title)
        self.pages = pages

    def _compute_change_extent(self) -> ChangeExtent:
        # TODO: with some more infrastructure, we can track what pages
        # contributed the links, and compute something better
        return ChangeExtent.ALL


class LinksTagPage(TemplatePage, AutoPage):
    """
    Page with an autogenerated link collection from a link tag.
    """

    TYPE = "links_tag"
    TEMPLATE = "data-links.html"

    def __init__(self, *args: Any, link_collection: LinkCollection, **kw: Any):
        super().__init__(*args, **kw)
        self.syndicated = False
        self.link_collection = link_collection

    @property
    def src_abspath(self) -> str | None:
        return None

    def _compute_change_extent(self) -> ChangeExtent:
        # TODO: with some more infrastructure, we can track what pages
        # contributed the links, and compute something better
        return ChangeExtent.ALL



from unittest import TestCase

from staticsite import toposort


class TestTopoSort(TestCase):
    def test_nodeps(self):
        graph = {}
        graph["a"] = set()
        graph["b"] = set()
        graph["c"] = set()
        self.assertEqual(toposort.sort(graph), ["a", "b", "c"])

    def test_singledeps(self):
        graph = {}
        graph["a"] = set()
        graph["a"] = {"b"}
        graph["b"] = {"c"}
        self.assertEqual(toposort.sort(graph), ["c", "b", "a"])

    def test_cycle(self):
        graph = {}
        graph["b"] = {"a"}
        graph["c"] = {"b"}
        graph["a"] = {"c"}
        with self.assertRaises(ValueError) as e:
            self.assertEqual(toposort.sort(graph), ["a", "c", "b"])
        self.assertEqual(
            str(e.exception), "('nodes are in a cycle', ['b', 'c', 'a', 'b'])"
        )



from __future__ import annotations

import asyncio
import gc
import json
import logging
import mimetypes
import os
from collections.abc import Awaitable
from typing import TYPE_CHECKING, cast

import pyinotify
import tornado.httpserver
import tornado.ioloop
import tornado.netutil
import tornado.web
import tornado.websocket
from tornado.web import url

from staticsite.site import Site
from staticsite.utils import timings

from .pagefs import PageFS

if TYPE_CHECKING:
    from staticsite.settings import Settings


log = logging.getLogger("serve")


class ChangeMonitor:
    """
    Trigger an event when a file is removed, then recreate the file
    """

    def __init__(self, app: Application):
        self.app = app
        self.loop = asyncio.get_event_loop()
        # Set up pyinotify.
        # See https://stackoverflow.com/questions/26414052/watch-for-a-file-with-asyncio
        self.watch_manager = pyinotify.WatchManager()
        # Map absolute paths to Watch handlers
        self.watches: dict[str, dict[str, int]] = {}
        # self.watch = self.watch_manager.add_watch(self.media_dir, pyinotify.IN_DELETE)
        self.notifier = pyinotify.AsyncioNotifier(
            self.watch_manager, self.loop, default_proc_fun=self.on_event
        )
        # Pending trigger
        self.pending: asyncio.TimerHandle | None = None

    def update_watch_dirs(self, dirs: list[str]) -> None:
        dirs = [os.path.realpath(d) for d in dirs]

        for path in self.watches.keys() - dirs:
            watch = self.watches.pop(path)
            self.watch_manager.rm_watch(list(watch.values()))
            log.info("%s: removing watch", path)

        for path in set(dirs) - self.watches.keys():
            self.watches[path] = self.watch_manager.add_watch(
                path, pyinotify.IN_CLOSE_WRITE | pyinotify.IN_DELETE, rec=True
            )
            log.info("%s: adding watch", path)

    def notify(self) -> None:
        self.pending = None
        self.app.trigger_reload()

    def on_event(self, event: pyinotify.Event) -> None:
        """
        Handle incoming asyncio events
        """
        if event.name.startswith(".") and event.name != ".staticsite":
            return

        if os.path.basename(event.path).startswith("."):
            return

        # Check that it's not an event from inside a hidden directory like
        # .staticsite-cache
        event_path = os.path.realpath(event.path)
        for path in self.watches.keys():
            if event_path == path:
                break
            if event_path.startswith(path):
                relpath = os.path.relpath(event_path, path)
                while relpath:
                    dirname, basename = os.path.split(relpath)
                    if basename.startswith("."):
                        return
                    relpath = dirname
                break
        else:
            # Event not for a path that we watch
            return

        log.debug("Received event %r", event)

        # Introduce a delay of 0.1s from the last event received, to notify
        # only once in case of a burst of events
        if self.pending is not None:
            self.pending.cancel()
            self.pending = None
        self.pending = self.loop.call_later(0.1, self.notify)


class PageSocket(tornado.websocket.WebSocketHandler):
    def open(self, *args: str, **kwargs: str) -> Awaitable[None] | None:
        log.debug("WebSocket connection opened")
        cast(Application, self.application).add_page_socket(self)
        return None

    def on_message(self, message: str | bytes) -> Awaitable[None] | None:
        log.debug("WebSocket message received: %r", message)
        return None

    def on_close(self) -> None:
        log.debug("WebSocket connection closed")
        cast(Application, self.application).remove_page_socket(self)


class ServePage(tornado.web.RequestHandler):
    def get(self) -> None:
        if self.request.protocol == "http":
            self.ws_url = (
                "ws://"
                + self.request.host
                + self.application.reverse_url("page_socket")
            )
        else:
            self.ws_url = (
                "wss://"
                + self.request.host
                + self.application.reverse_url("page_socket")
            )

        relpath, content = cast(Application, self.application).pages.render(
            self.request.path, server=self.application, handler=self
        )
        if relpath is None:
            self.send_error(404)
        else:
            self.set_header("Content-Type", mimetypes.guess_type(relpath)[0])
            self.finish(content)


class Application(tornado.web.Application):
    def __init__(self, settings: Settings):
        mimetypes.init()

        urls = [
            url(r"/_server/websocket", PageSocket, name="page_socket"),
            url(r".*", ServePage, name="page"),
        ]

        super().__init__(
            # TODO: this seems like a glitch in tornado's typing
            urls,  # type: ignore
            xsrf_cookies=True,
        )

        self.site_settings: Settings = settings
        self.site: Site
        self.pages: PageFS
        self.page_sockets: set[PageSocket] = set()
        self.change_monitor = ChangeMonitor(self)

    def add_page_socket(self, handler: PageSocket) -> None:
        self.page_sockets.add(handler)

    def remove_page_socket(self, handler: PageSocket) -> None:
        self.page_sockets.discard(handler)

    def trigger_reload(self) -> None:
        log.info("Content change detected: reloading site")
        self.reload()
        payload = json.dumps({"event": "reload"})
        for handler in self.page_sockets:
            handler.write_message(payload)

    def reload(self) -> None:
        # (re)instantiate site
        # FIXME: do the build in a thread worker?
        self.site = Site(settings=self.site_settings)
        with timings("Loaded site in %fs"):
            self.site.load()

        self.change_monitor.update_watch_dirs(self.get_source_dirs())

        self.pages = PageFS(self.site)
        gc.collect()

    def get_source_dirs(self) -> list[str]:
        """
        Get the directories used as sources to the site
        """
        # TODO: reload when source changes
        # See /usr/lib/python3/dist-packages/tornado/autoreload.py:_reload_on_update
        # for example code on how to list source files for the running program
        dirs = [self.site.content_root]
        dirs.extend(self.site.theme.feature_dirs)
        dirs.extend(self.site.theme.template_lookup_paths)
        for name in self.site.theme.system_assets:
            dirs.append(os.path.join("/usr/share/javascript", name))
        return dirs



from __future__ import annotations

import argparse
import logging
import os
from typing import Any

from ..command import Command, Fail, SiteCommand, register

log = logging.getLogger("serve")


class ServerMixin(Command):
    def start_server(self) -> None:
        try:
            import tornado.httpserver
            import tornado.netutil
        except ModuleNotFoundError:
            raise Fail("python3-tornado is not installed")
        import asyncio
        import webbrowser

        from .server import Application

        sockets = tornado.netutil.bind_sockets(self.args.port, self.args.host)
        pairs = []
        for s in sockets:
            pairs.append(s.getsockname()[:2])
        pairs.sort()
        host, port = pairs[0]

        if ":" in host:
            host = f"[{host}]"
        url = f"http://{host}:{port}"
        log.info("Serving on %s", url)

        if not hasattr(self.settings, "SITE_URL"):
            self.settings.SITE_URL = url

        app = Application(self.settings)
        app.reload()

        server = tornado.httpserver.HTTPServer(app)
        server.add_sockets(sockets)

        async def open_browser() -> None:
            webbrowser.open_new_tab(url)

        if not getattr(self.args, "no_start", True):
            tornado.ioloop.IOLoop.current().add_callback(open_browser)
        else:
            print(url)

        loop = asyncio.get_event_loop()
        loop.run_forever()

    def run(self) -> None:
        self.start_server()


@register
class Serve(ServerMixin, SiteCommand):
    "Serve the site over HTTP, building it in memory on demand"

    @classmethod
    def add_subparser(
        cls, subparsers: argparse._SubParsersAction[Any]
    ) -> argparse.ArgumentParser:
        parser = super().add_subparser(subparsers)
        parser.add_argument(
            "--port",
            "-p",
            action="store",
            type=int,
            default=8000,
            help="port to use (default: 8000)",
        )
        parser.add_argument(
            "--host",
            action="store",
            type=str,
            default="localhost",
            help="host to bind to (default: localhost)",
        )
        return parser


@register
class Show(ServerMixin, Command):
    "Show the current directory in a browser"

    def __init__(self, *args: Any, **kw: Any) -> None:
        super().__init__(*args, **kw)

        # Set default project root if undefined
        self.settings.PROJECT_ROOT = self.args.dir
        if self.settings.PROJECT_ROOT is None:
            self.settings.PROJECT_ROOT = os.getcwd()

        # Command line overrides for settings
        if self.args.theme:
            self.settings.THEME = (os.path.abspath(self.args.theme),)
        if self.args.draft:
            self.settings.DRAFT_MODE = True

        # Do not clutter previewed directories with .staticsite-cache directories
        self.settings.CACHE_REBUILDS = False

    @classmethod
    def add_subparser(
        cls, subparsers: argparse._SubParsersAction[Any]
    ) -> argparse.ArgumentParser:
        parser = super().add_subparser(subparsers)

        parser.add_argument(
            "dir", nargs="?", help="directory to show (default: the current directory)"
        )
        parser.add_argument(
            "--theme", help="theme directory location. Overrides settings.THEME"
        )
        parser.add_argument(
            "--draft",
            action="store_true",
            help="do not ignore pages with date in the future",
        )
        parser.add_argument(
            "--no-start",
            "-n",
            action="store_true",
            help="do not start a browser automatically, print the URL instead",
        )
        parser.add_argument(
            "--port",
            "-p",
            action="store",
            type=int,
            default=0,
            help="port to use (default: randomly allocated)",
        )
        parser.add_argument(
            "--host",
            action="store",
            type=str,
            default="localhost",
            help="host to bind to (default: localhost)",
        )
        return parser



from __future__ import annotations

import collections.abc
import datetime
import enum
import logging
import os
from collections.abc import Iterator, Sequence
from functools import cached_property
from typing import TYPE_CHECKING, Any, Optional, TypeVar, Union, cast, overload

import jinja2
import markupsafe

from . import fields
from .render import RenderedString
from .site import Path, SiteElement
from .utils.arrange import arrange

if TYPE_CHECKING:
    from .file import File
    from .node import Node
    from .render import RenderedElement
    from .site import Site

log = logging.getLogger("page")

T = TypeVar("T")


class ChangeExtent(enum.IntEnum):
    """
    What kind of changes happened on this page since last build
    """

    # Page is unchanged
    UNCHANGED = 0
    # Page changed in contents but not in metadata
    CONTENTS = 1
    # Page changed completely
    ALL = 2


class PageNotFoundError(Exception):
    pass


class PageValidationError(Exception):
    def __init__(self, page: Page, msg: str):
        self.page = page
        self.msg = msg


class PageMissesFieldError(PageValidationError):
    def __init__(self, page: Page, field: str):
        super().__init__(page, f"missing required field meta.{field}")


P = TypeVar("P", bound="Page")
P1 = TypeVar("P1", bound="Page")
V = TypeVar("V")


class CrossreferenceField(fields.Field[P, V]):
    """
    Call crossreference() on the page when this field is set
    """

    def __set__(self, page: P, value: Any) -> None:
        page.site.pages_to_crossreference.add(page)
        super().__set__(page, value)


class CreatedFromField(fields.Const[P, Optional["Page"]]):
    """
    Field containing a Page
    """

    def _clean(self, page: P, value: Any) -> Page | None:
        if value is None:
            return None
        elif isinstance(value, Page):
            return value
        else:
            raise TypeError(
                f"invalid value of type {type(value)} for {page!r}.{self.name}:"
                " expecting, None or Page"
            )


class ConstPageField(fields.ConstTypeField[P, P1]):
    """
    Field containing a string
    """

    def __init__(self, *, cls: type[P1], **kw: Any):
        super().__init__(cls=cls, **kw)


class Pages(collections.abc.Sequence["Page"]):
    """
    List of pages, resolved at `crossreference` time, that can be provided as:
     - path to another page, relative to this page
     - page filter
    """

    def __init__(self, page: Page, value: Any):
        # Reference page for resolving relative paths
        self.page = page

        # Resolved list of pages
        self.pages: list[Page] | None = None

        # Filter expression to enumerate pages
        self.query: dict[str, Any] | None = None

        if isinstance(value, str):
            self.query = {"path": value}
        elif isinstance(value, list):
            self.pages = value
        elif isinstance(value, dict):
            self.query = value
        elif isinstance(value, Pages):
            self.page = value.page
            self.pages = value.pages
            self.query = value.query
        else:
            raise RuntimeError("pages field is not string, list of pages, or dict")

    def resolve(self) -> None:
        """
        Fill self.pages using self.query
        """
        if self.pages is not None:
            return

        if self.query is None:
            self.pages = []
            return

        # Replace the dict with the expanded list of pages
        # Do not include self in the result list
        self.pages = [p for p in self.page.find_pages(**self.query) if p != self.page]

    def to_dict(self) -> list[Page] | None:
        return self.pages

    def __repr__(self) -> str:
        if self.pages is None:
            return repr(self.query)
        else:
            return repr(self.pages)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, list):
            return self.pages == other
        elif isinstance(other, dict):
            return self.query == other
        elif isinstance(other, Pages):
            return (
                self.page == other.page
                and self.query == other.query
                or self.pages == other.pages
            )
        else:
            return False

    def __len__(self) -> int:
        if self.pages is None:
            raise RuntimeError(
                f"{self.page}.pages is accessed before the crossreference step has run"
            )
        return len(self.pages)

    @overload
    def __getitem__(self, key: int, /) -> Page:
        ...

    @overload
    def __getitem__(self, key: slice, /) -> Sequence[Page]:
        ...

    def __getitem__(self, key: int | slice, /) -> Page | Sequence[Page]:
        if self.pages is None:
            raise RuntimeError(
                f"{self.page}.pages is accessed before the crossreference step has run"
            )
        return self.pages.__getitem__(key)

    def __iter__(self) -> Iterator[Page]:
        if self.pages is None:
            raise RuntimeError(
                f"{self.page}.pages is accessed before the crossreference step has run"
            )
        return self.pages.__iter__()

    def __reversed__(self) -> Iterator[Page]:
        if self.pages is None:
            raise RuntimeError(
                f"{self.page}.pages is accessed before the crossreference step has run"
            )
        return self.pages.__reversed__()

    def arrange(self, sort: str, limit: int | None = None) -> list[Page]:
        """
        Sort the pages by ``sort`` and take the first ``limit`` ones
        """
        if self.pages is None:
            raise RuntimeError(
                f"{self.page}.arrange is accessed before the crossreference step has run"
            )
        return arrange(self.pages, sort=sort, limit=limit)


class PagesField(CrossreferenceField["Page", Pages]):
    """
    The `pages` metadata can use to select a set of pages shown by the current
    page. Although default `page.html` template will not do anything with them,
    other page templates, like `blog.html`, use this to select the pages to show.

    The `pages` field allows defining a [page filter](../page-filter.md) which
    will be replaced with a list of matching pages.

    To select pages, the `pages` metadata is set to a dictionary that select pages
    in the site, with the `path`, and taxonomy names arguments similar to the
    `site_pages` function in [templates](../templates.md).

    See [Selecting pages](../page-filter.md) for details.

    ## Example blog page

    ```jinja2
    {% extends "base.html" %}

    {% block front_matter %}
    ---
    pages:
      path: blog/*
      limit: 10
      sort: "-date"
    syndication:
      add_to:
        path: blog/*
      title: "Enrico's blog feed"
    {% endblock %}

    {% import 'lib/blog.html' as blog %}

    {% block title %}Enrico's blog{% endblock %}

    {% block nav %}
    {{super()}}
    <li class="nav-item"><a class="nav-link" href="{{ url_for('/blog/archive.html') }}">Archive</a></li>
    {% endblock %}

    {% block content %}

    <h1 class="display-4">Last 10 blog posts</h1>

    {{blog.pages(page)}}

    {% endblock %}
    ```
    """

    def _clean(self, page: Page, value: Any) -> Pages:
        return Pages(page, value)


class TemplateField(fields.Field["TemplatePage", Union[str, jinja2.Template]]):
    """
    Template name or compiled template, taking its default value from Page.TEMPLATE
    """

    def __get__(
        self, page: TemplatePage, type: type[TemplatePage] | None = None
    ) -> str | jinja2.Template:
        if self.name not in page.__dict__:
            page.__dict__[self.name] = page.TEMPLATE
            return page.TEMPLATE
        else:
            return cast(str, page.__dict__[self.name])

    def _clean(self, obj: Page, value: Any) -> str | jinja2.Template:
        if isinstance(value, str):
            return value
        elif isinstance(value, jinja2.Template):
            return value
        else:
            raise TypeError(
                f"invalid value of type {type(value)} for {obj!r}.{self.name}:"
                " expecting str or jinja2.Template"
            )


class PageDate(fields.Date["Page"]):
    """
    Make sure, on page load, that the element is a valid aware datetime object
    """

    def __get__(self, page: Page, type: type[Page] | None = None) -> datetime.datetime:
        if (date := page.__dict__.get(self.name)) is None:
            if (src := getattr(page, "src", None)) is not None and src.stat is not None:
                date = page.site.localized_timestamp(src.stat.st_mtime)
            else:
                date = page.site.generation_time
            page.__dict__[self.name] = date
        return date


class Draft(fields.Bool["SourcePage"]):
    """
    Make sure the draft exists and is a bool, computed according to the date
    """

    def __get__(self, page: SourcePage, type: type[SourcePage] | None = None) -> Any:
        if (value := page.__dict__.get(self.name)) is None:
            value = page.date > page.site.generation_time
            page.__dict__[self.name] = value
            return value
        else:
            return value


class RenderedField(fields.Str["Page"]):
    """
    Field whose value is rendered from other fields
    """

    def __get__(self, page: Page, type: type[Page] | None = None) -> str | None:
        if (cur := page.__dict__.get(self.name)) is None:
            value: str
            if tpl := getattr(page, "template_" + self.name, None):
                # If a template exists, render it
                value = markupsafe.Markup(tpl.render(page=page))
                self.__dict__[self.name] = value
                return value
            elif self.default is not None:
                self.__dict__[self.name] = self.default
                return self.default
            else:
                return None
        else:
            return cast(str, cur)


class RenderedTitleField(fields.Str["Page"]):
    """
    Render the tile for a page, defaulting to site_name if missing
    """

    def __get__(self, page: Page, type: type[Page] | None = None) -> str:
        if (cur := page.__dict__.get(self.name)) is None:
            value: str
            if tpl := getattr(page, "template_" + self.name, None):
                # If a template exists, render it
                value = markupsafe.Markup(tpl.render(page=page))
            elif page.site_name is None:
                raise RuntimeError("site_name not set")
            else:
                value = page.site_name
            self.__dict__[self.name] = value
            return value
        else:
            return cast(str, cur)


class Related(collections.abc.MutableMapping[str, "Page"]):
    """
    Container for related pages
    """

    def __init__(self, page: Page):
        self.page = page
        self.pages: dict[str, str | Page] = {}

    def __repr__(self) -> str:
        return f"Related({self.page!r}, {self.pages!r}))"

    def __str__(self) -> str:
        return self.pages.__str__()

    def __getitem__(self, name: str, /) -> Page:
        val = self.pages[name]

        if isinstance(val, str):
            page = self.page.resolve_path(val)
            self.pages[name] = page
            return page
        else:
            return val

    def __setitem__(self, name: str, page: str | Page) -> None:
        if (old := self.pages.get(name)) is not None and old != self.page:
            log.warning(
                "%s: attempt to set related.%s to %r but it was already %r",
                self,
                name,
                page,
                old,
            )
            return
        self.pages[name] = page

    def __delitem__(self, name: str, /) -> None:
        self.pages.__delitem__(name)

    def __len__(self) -> int:
        return self.pages.__len__()

    def __iter__(self) -> Iterator[str]:
        return self.pages.__iter__()

    def to_dict(self) -> dict[str, Any]:
        return self.pages

    @overload
    def get(self, key: str, /) -> Page | None:
        ...

    @overload
    def get(self, key: str, /, default: Page | T) -> Page | T:
        ...

    def get(self, key: str, default: Page | T | None = None) -> Page | T | None:
        if (val := self.pages.get(key, None)) is None:
            return default
        elif isinstance(val, str):
            page = self.page.resolve_path(val)
            self.pages[key] = page
            return page
        else:
            return val

    def _ensure_all_resolved(self) -> None:
        for name, page in self.pages.items():
            if isinstance(page, str):
                self.pages[name] = self.page.resolve_path(page)

    def keys(self) -> collections.abc.KeysView[str]:
        return self.pages.keys()

    def values(self) -> collections.abc.ValuesView[Page]:
        self._ensure_all_resolved()
        return cast(collections.abc.ValuesView[Page], self.pages.values())

    def items(self) -> collections.abc.ItemsView[str, Page]:
        self._ensure_all_resolved()
        return cast(collections.abc.ItemsView[str, Page], self.pages.items())

    def __eq__(self, obj: Any) -> bool:
        if isinstance(obj, Related):
            return self.page == obj.page and self.pages == obj.pages
        elif isinstance(obj, dict):
            return self.pages == obj
        elif obj is None:
            return False
        else:
            return False


class RelatedField(fields.Field["Page", Related]):
    """
    Readonly mapping of pages related to this page, indexed by name.

    If there are no related pages, `page.meta.related` will be guaranteed to exist
    as an empty dictionary.

    Features can add to this. For example, [syndication](syndication.md) can add
    `meta.related.archive`, `meta.related.rss`, and `meta.related.atom`.
    """

    def __get__(self, page: Page, type: type[Page] | None = None) -> Related:
        if (value := page.__dict__.get(self.name)) is None:
            value = Related(page)
            page.__dict__[self.name] = value
        return value

    def __set__(self, page: Page, value: Related) -> None:
        related = self.__get__(page)
        related.pages.update(value)


class Meta:
    """
    Read-only dict accessor to Page's fields
    """

    def __init__(self, page: Page):
        self._page = page

    def __getitem__(self, key: str) -> Any:
        if (field := self._page._fields.get(key)) is None:
            raise KeyError(key)

        if field.internal:
            raise KeyError(key)

        try:
            res = getattr(self._page, key)
            # print(f"{self._page!r} meta[{key!r}] = {res!r}")
            return res
        except AttributeError:
            raise KeyError(key)

    def __contains__(self, key: str) -> bool:
        if (field := self._page._fields.get(key)) is None:
            return False
        if field.internal:
            return False

        return getattr(self._page, key) is not None

    def get(self, key: str, default: Any = None) -> Any:
        if (field := self._page._fields.get(key)) is None:
            return default
        if field.internal:
            return default

        return getattr(self._page, key, default)

    def to_dict(self) -> dict[str, Any]:
        """
        Return a dict with all the values of this Meta, including the inherited
        ones
        """
        res = {}
        for key, field in self._page._fields.items():
            if field.internal:
                continue
            if (val := getattr(self._page, key)) is not None:
                res[key] = val
        return res


class Page(SiteElement):
    """
    A source page in the site.

    This can be a static asset, a file to be rendered, a taxonomy, a
    directory listing, or anything else.
    """

    # Page type
    TYPE: str

    date = PageDate(
        doc="""
        Publication date for the page.

        A python datetime object, timezone aware. If the date is in the future when
        `ssite` runs, the page will be consider a draft and will be ignored. Use `ssite
        --draft` to also consider draft pages.

        If missing, the modification time of the file is used.
    """
    )

    copyright = RenderedField(
        doc="""
        Copyright notice for the page. If missing, it's generated using
        `template_copyright`.
    """
    )

    title = RenderedTitleField(
        doc="""
        Page title.

        If missing:

         * it is generated via template_title, if present
         * the first title found in the page contents is used.
         * in the case of jinaj2 template pages, the contents of `{% block title %}`,
           if present, is rendered and used.
         * if the page has no title, the title of directory indices above this page is
           inherited.
         * if still no title can be found, the site name is used as a default.
    """
    )

    description = RenderedField(
        doc="""
        The page description. If omitted, the page will have no description.
    """
    )

    indexed = fields.Bool["Page"](
        default=False,
        doc="""
        If true, the page appears in [directory indices](dir.md) and in
        [page filter results](page_filter.md).

        It defaults to true at least for [Markdown](markdown.md),
        [reStructuredText](rst.rst), and [data](data.md) pages.
    """,
    )

    pages = PagesField(structure=True)

    related = RelatedField(structure=True)

    def __init__(
        self,
        site: Site,
        *,
        node: Node,
        search_root_node: Node,
        dst: str,
        leaf: bool,
        directory_index: bool = False,
        **kw: Any,
    ) -> None:
        # Set these fields early as they are used by __str__/__repr__

        # Node where this page is installed in the rendered structure
        self.node: Node = node

        # Set to True if this page is shown as a file url, False if it's shown
        # as a path url
        self.leaf: bool = leaf

        # Name of the file used to render this page on build
        self.dst: str = dst

        super().__init__(site, **kw)

        # Read-only, dict-like accessor to the page's fields
        self.meta: Meta = Meta(self)
        # Node to use as initial node for find_pages
        self.search_root_node: Node = search_root_node
        # Set to True if this page is a directory index. This affects the root
        # of page lookups relative to this page
        self.directory_index: bool = directory_index
        # Basename of this page in the source tree, or None if it's an
        # autogenerated page
        self.source_name: str | None = None

        # Check the existence of other mandatory fields
        if self.site_url is None:
            raise PageMissesFieldError(self, "site_url")

    @property
    def site_path(self) -> str:
        """
        Accessor to support the migration away from meta['build_path']
        """
        # return self.meta["build_path"]
        if self.leaf:
            return os.path.join(self.node.path, self.dst)
        else:
            return self.node.path

    def add_related(self, name: str, page: Page) -> None:
        """
        Set the page as meta.related.name
        """
        self.related[name] = page

    def find_pages(
        self,
        path: str | None = None,
        limit: int | None = None,
        sort: str | None = None,
        **kw: Any,
    ) -> list[Page]:
        """
        If not set, default root to the path of the containing directory for
        this page
        """
        from .page_filter import PageFilter

        f = PageFilter(
            self.site,
            path=path,
            limit=limit,
            sort=sort,
            root=self.search_root_node,
            **kw,
        )
        return f.filter()

    def lookup_page(self, path: Path) -> Page | None:
        return self.search_root_node.lookup_page(path)

    def url_for(
        self, target: str | Page, absolute: bool = False, static: bool = False
    ) -> str:
        """
        Generate a URL for a page, specified by:

        * an absolute path
        * a path relative to this page
        * the target page itself
        """
        # print(f"Page.url_for {self=!r}, {target=!r}")
        page: Page
        if isinstance(target, str):
            page = self.resolve_path(target, static=static)
        else:
            page = target

        # print(f"Page.url_for {self=!r}, {target=!r}, {page=!r}")

        # If the destination has a different site_url, generate an absolute url
        if self.site_url != page.site_url:
            absolute = True

        if self.site.root.site_path:
            path = os.path.join(self.site.root.site_path, page.site_path).strip("/")
        else:
            path = page.site_path.strip("/")

        if absolute:
            if page.site_url is None:
                raise RuntimeError("site_url is None")
            site_url = page.site_url.rstrip("/")
            return f"{site_url}/{path}"
        else:
            return "/" + path

    def crossreference(self) -> None:
        """
        Called at the beginning of the 'crossreference' step when the page is in
        site.pages_to_crossreference
        """
        if self.pages is not None:
            self.pages.resolve()

        if self.pages:
            # Update the page date to the max of the pages dates
            max_date = max(p.date for p in self.pages)
            self.date = max(max_date, self.date)

    def render(self, **kw: Any) -> RenderedElement:
        """
        Return a RenderedElement that can produce the built version of this page
        """
        raise NotImplementedError(f"{self.__class__.__name__}.render not implemented")

    def check(self) -> None:
        pass

    def __str__(self) -> str:
        return self.site_path

    def __repr__(self) -> str:
        return f"{self.TYPE}:auto:{self.site_path}"

    def to_dict(self) -> dict[str, Any]:
        from .utils import dump_meta

        res = {
            "meta": dump_meta(self.meta),
            "type": self.TYPE,
            "site_path": self.site_path,
            "build_path": os.path.join(self.node.path, self.dst),
        }
        return res

    def _compute_change_extent(self) -> ChangeExtent:
        """
        Check how much this page has changed since the last build
        """
        return ChangeExtent.UNCHANGED

    @cached_property
    def change_extent(self) -> ChangeExtent:
        """
        How much this page has changed since the last build
        """
        if self.site.last_load_step < self.site.LOAD_STEP_CROSSREFERENCE:
            raise RuntimeError(
                "Page.change_extent referenced before running page crossreference stage"
            )
        return self._compute_change_extent()


class SourcePage(Page):
    """
    Page loaded from site sources
    """

    draft = Draft(
        doc="""
If true, the page is still a draft and will not appear in the destination site,
unless draft mode is enabled.

It defaults to false, or true if `meta.date` is in the future.
"""
    )

    def __init__(
        self,
        site: Site,
        *,
        node: Node,
        src: File,
        old_footprint: dict[str, Any] | None = None,
        **kw: Any,
    ):
        # Information about the source file for this page
        # Set right away so that __repr__ works
        self.src: File = src
        super().__init__(site, parent=node, node=node, **kw)
        self.source_name: str = os.path.basename(self.src.relpath)

        # Cached footprint from the previous run, or None
        self.old_footprint: dict[str, Any] | None = old_footprint

    def __str__(self) -> str:
        return self.site_path

    def __repr__(self) -> str:
        return f"{self.TYPE}:{self.src.relpath}"

    def to_dict(self) -> dict[str, Any]:
        res = super().to_dict()
        res["src"] = {
            "relpath": str(self.src.relpath),
            "abspath": str(self.src.abspath),
        }
        return res

    def _compute_footprint(self) -> dict[str, Any]:
        """
        Return a dict with information that can be used to evaluate changes in
        incremental builds
        """
        if self.site.last_load_step < self.site.LOAD_STEP_CROSSREFERENCE:
            raise RuntimeError(
                "SourcePage._compute_footprint referenced before running page crossreference stage"
            )
        res: dict[str, Any] = {
            "mtime": self.src.stat.st_mtime,
            "size": self.src.stat.st_size,
        }
        # We can cast to list[Page] since we made sure we only run after the crossreference stage
        if pages := self.pages:
            res["pages"] = [
                page.src.relpath for page in pages if isinstance(page, SourcePage)
            ]
        return res

    @cached_property
    def footprint(self) -> dict[str, Any]:
        """
        Dict with information that can be used to evaluate changes in
        incremental builds
        """
        return self._compute_footprint()

    def _compute_change_extent(self) -> ChangeExtent:
        res = super()._compute_change_extent()
        if (old := self.old_footprint) is None:
            return ChangeExtent.ALL
        if (
            old.get("mtime") < self.footprint["mtime"]
            or old.get("size") != self.footprint["size"]
        ):
            return ChangeExtent.ALL
        if res == ChangeExtent.UNCHANGED and self.footprint and self.old_footprint:
            if set(self.footprint.get("pages", ())) != set(
                self.old_footprint.get("pages", ())
            ):
                return ChangeExtent.ALL
        return res


class StandaloneAutoPage(Page):
    """
    Page that has been autogenerated, but not from another page
    """

    def __init__(self, site: Site, *, node: Node, **kw: Any) -> None:
        super().__init__(site, parent=node, node=node, **kw)


class AutoPage(Page):
    """
    Autogenerated page
    """

    created_from = CreatedFromField["AutoPage"](
        doc="""
        Page that generated this page.

        This is only set in autogenerated pages.
    """
    )

    def __init__(self, site: Site, *, created_from: Page, **kw: Any) -> None:
        if created_from is None:
            raise RuntimeError("created_from is None in AutoPage")
        super().__init__(site, parent=created_from, created_from=created_from, **kw)


class FrontMatterPage(SourcePage):
    """
    Page with a front matter in its sources
    """

    front_matter = fields.Dict["FrontMatterPage"](
        internal=True,
        doc="""
        Front matter as parsed by the source file
    """,
    )

    def _compute_footprint(self) -> dict[str, Any]:
        res = super()._compute_footprint()
        fm = {}
        for k, v in self.front_matter.items():
            if isinstance(v, datetime.datetime):
                fm[k] = v.strftime("%Y-%m-%d %H:%M:%S %Z")
            else:
                fm[k] = v
        res["fm"] = fm
        return res

    def _compute_change_extent(self) -> ChangeExtent:
        res = super()._compute_change_extent()
        if res == ChangeExtent.UNCHANGED:
            return res
        if self.old_footprint is not None and self.footprint[
            "fm"
        ] == self.old_footprint.get("fm"):
            return ChangeExtent.CONTENTS
        else:
            return ChangeExtent.ALL


class TemplatePage(Page):
    """
    Page that renders using a Jinja2 template
    """

    # Default page template
    TEMPLATE = "page.html"

    template = TemplateField(
        doc="""
        Template used to render the page. Defaults to `page.html`, although specific
        pages of some features can default to other template names.

        Use this similarly to [Jekill's layouts](https://jekyllrb.com/docs/step-by-step/04-layouts/).
    """
    )

    @cached_property
    def page_template(self) -> jinja2.Template:
        template = self.meta["template"]
        if isinstance(template, jinja2.Template):
            return template
        return self.site.theme.jinja2.get_template(template)

    @jinja2.pass_context
    def html_body(self, context: jinja2.runtime.Context, **kw: Any) -> str:
        """
        Render the full body of the page, with UI elements excluding
        navigation, headers, footers
        """
        return ""

    @jinja2.pass_context
    def html_inline(self, context: jinja2.runtime.Context, **kw: Any) -> str:
        """
        Render the content of the page to be shown inline, like in a blog page.

        Above-the-fold content only, small images, no UI elements
        """
        return ""

    @jinja2.pass_context
    def html_feed(self, context: jinja2.runtime.Context, **kw: Any) -> str:
        """
        Render the content of the page to be shown in a RSS/Atom feed.

        It shows the whole contents, without any UI elements, and with absolute
        URLs
        """
        return ""

    def render(self, **kw: Any) -> RenderedElement:
        """
        Render the full page, from the <html> tag downwards.
        """
        kw["render_style"] = "full"
        rendered = self.render_template(self.page_template, template_args=kw)
        return RenderedString(rendered)

    def render_template(
        self, template: jinja2.Template, template_args: dict[Any, Any] | None = None
    ) -> str:
        """
        Render a jinja2 template, logging things if something goes wrong
        """
        if template_args is None:
            template_args = {}
        template_args["page"] = self
        # try:
        return template.render(**template_args)
        # except jinja2.TemplateError as e:
        #     log.error("%s: failed to render %s: %s",
        #               template.filename, self.src.relpath if self.src else repr(self), e)
        #     log.debug("%s: failed to render %s: %s",
        #               template.filename, self.src.relpath if self.src else repr(self), e, exc_info=True)
        #     # TODO: return a "render error" page? But that risks silent errors
        #     return None


class ImagePage(Page):
    width = fields.Int["ImagePage"](
        doc="""
        Image width
    """
    )
    height = fields.Int["ImagePage"](
        doc="""
        Image height
    """
    )

    def get_img_attributes(
        self, type: str | None = None, absolute: bool = False
    ) -> dict[str, str]:
        """
        Given a path to an image page, return a dict with <img> attributes that
        can be used to refer to it
        """
        res = {
            "alt": self.title,
        }

        if type is not None:
            # If a specific version is required, do not use srcset
            rel = self.related.get(type, self)
            if not isinstance(rel, ImagePage):
                raise RuntimeError(
                    f"{self}: related page of type {type!r} has type {rel.__class__.__name__} instead of ImagePage"
                )
            res["width"] = str(rel.width)
            res["height"] = str(rel.height)
            res["src"] = self.url_for(rel, absolute=absolute)
        else:
            # https://developers.google.com/web/ilt/pwa/lab-responsive-images
            # https://developer.mozilla.org/en-US/docs/Learn/HTML/Multimedia_and_embedding/Responsive_images
            srcsets = []
            for rel in self.related.values():
                if not isinstance(rel, ImagePage):
                    continue

                if (width := rel.width) is None:
                    continue

                url = self.url_for(rel, absolute=absolute)
                srcsets.append(f"{markupsafe.escape(url)} {width}w")

            if srcsets:
                width = self.width
                srcsets.append(f"{markupsafe.escape(self.url_for(self))} {width}w")
                res["srcset"] = ", ".join(srcsets)
                res["src"] = self.url_for(self, absolute=absolute)
            else:
                res["width"] = str(self.width)
                res["height"] = str(self.height)
                res["src"] = self.url_for(self, absolute=absolute)

        return res



from __future__ import annotations

import contextlib
import datetime
import logging
import os
import time
from collections.abc import Generator
from typing import Any

import pytz

log = logging.getLogger("utils")


def format_date_rfc822(dt: datetime.datetime) -> str:
    from email.utils import formatdate

    return formatdate(dt.timestamp())


def format_date_rfc3339(dt: datetime.datetime) -> str:
    dt = dt.astimezone(pytz.utc)
    return dt.strftime("%Y-%m-%dT%H:%M:%SZ")


def format_date_w3cdtf(dt: datetime.datetime) -> str:
    offset = dt.utcoffset()
    if offset:
        offset_sec = offset.days * 24 * 3600 + offset.seconds
        offset_hrs = offset_sec // 3600
        offset_min = offset_sec % 3600
        tz_str = f"{offset_hrs:+03d}:{offset_min // 60:02d}"
    else:
        tz_str = "Z"
    return dt.strftime("%Y-%m-%dT%H:%M:%S") + tz_str


def format_date_iso8601(dt: datetime.datetime) -> str:
    offset = dt.utcoffset()
    if offset:
        offset_sec = offset.days * 24 * 3600 + offset.seconds
        offset_hrs = offset_sec // 3600
        offset_min = offset_sec % 3600
        tz_str = f"{offset_hrs:+03d}:{offset_min // 60:02d}"
    else:
        tz_str = "Z"
    return dt.strftime("%Y-%m-%d %H:%M:%S") + tz_str


@contextlib.contextmanager
def timings(fmtstr: str, *args: Any, **kw: Any) -> Generator[None, None, None]:
    """
    Times the running of a command, and writes a log entry afterwards.

    The log entry is passed an extra command at the beginning with the elapsed
    time in floating point seconds.
    """
    start = time.perf_counter_ns()
    yield
    end = time.perf_counter_ns()
    log.info(fmtstr, (end - start) / 1_000_000_000, *args, extra=kw)


def dump_meta(
    val: Any,
) -> (
    None
    | bool
    | int
    | float
    | str
    | list[Any]
    | tuple[Any, ...]
    | set[Any]
    | dict[str, Any]
):
    """
    Dump data into a dict, for use with dump_meta in to_dict methods
    """
    import jinja2

    from ..page import Page

    if val is None:
        return None
    elif isinstance(val, (bool, int, float)):
        return val
    elif isinstance(val, str):
        return str(val)
    elif isinstance(val, Page):
        return f"{val.__class__.__name__}({val})"
    elif isinstance(val, dict):
        return {k: dump_meta(v) for k, v in val.items()}
    elif isinstance(val, (list, tuple, set)):
        return [dump_meta(v) for v in val]
    elif isinstance(val, jinja2.Template):
        return f"compiled:{val.name}"
    elif hasattr(val, "to_dict"):
        return dump_meta(val.to_dict())
    else:
        return str(val)


@contextlib.contextmanager
def open_dir_fd(path: str, dir_fd: int | None = None) -> Generator[int, None, None]:
    """
    Return a dir_fd for a directory. Supports dir_fd for opening.
    """
    res = os.open(path, os.O_RDONLY, dir_fd=dir_fd)
    try:
        yield res
    finally:
        os.close(res)



from __future__ import annotations

import os
from unittest import TestCase, mock

from staticsite.cmd.command import SiteCommand

from . import utils as test_utils


class TestBuildSettings(test_utils.MockSiteTestMixin, TestCase):
    """
    Test settings after command line parsing
    """

    def test_defaults(self):
        with self.site(test_utils.MockSite({}, auto_load_site=False)) as mocksite:
            with mock.patch("os.getcwd", return_value=mocksite.root):
                args = test_utils.Args()
                cmd = SiteCommand(args)
            settings = cmd.settings
            self.assertEqual(settings.PROJECT_ROOT, mocksite.root)
            self.assertEqual(settings.SITE_ROOT, "/")
            self.assertIsNone(settings.SITE_NAME)
            self.assertEqual(settings.ARCHETYPES, "archetypes")
            self.assertIsNone(settings.CONTENT)
            self.assertEqual(settings.THEME, "default")
            self.assertIsNone(settings.OUTPUT)
            self.assertIsNone(settings.TIMEZONE)
            self.assertEqual(settings.SYSTEM_ASSETS, [])
            self.assertFalse(settings.DRAFT_MODE)
            self.assertTrue(settings.CACHE_REBUILDS)

    def test_find_settings_py(self):
        files = {
            "settings.py": "SITE_NAME = 'Test Site'\n",
        }
        with self.site(test_utils.MockSite(files, auto_load_site=False)) as mocksite:
            with mock.patch("os.getcwd", return_value=mocksite.root):
                args = test_utils.Args()
                cmd = SiteCommand(args)
            settings = cmd.settings
            self.assertEqual(settings.PROJECT_ROOT, mocksite.root)
            self.assertEqual(settings.SITE_ROOT, "/")
            self.assertEqual(settings.SITE_NAME, "Test Site")
            self.assertEqual(settings.ARCHETYPES, "archetypes")
            self.assertIsNone(settings.CONTENT)
            self.assertEqual(settings.THEME, "default")
            self.assertIsNone(settings.OUTPUT)
            self.assertIsNone(settings.TIMEZONE)
            self.assertEqual(settings.SYSTEM_ASSETS, [])
            self.assertFalse(settings.DRAFT_MODE)
            self.assertTrue(settings.CACHE_REBUILDS)

    def test_find_dotstaticsite_py(self):
        files = {
            ".staticsite.py": "SITE_NAME = 'Test Site'\n",
        }
        with self.site(test_utils.MockSite(files, auto_load_site=False)) as mocksite:
            with mock.patch("os.getcwd", return_value=mocksite.root):
                args = test_utils.Args()
                cmd = SiteCommand(args)
            settings = cmd.settings
            self.assertEqual(settings.PROJECT_ROOT, mocksite.root)
            self.assertEqual(settings.SITE_ROOT, "/")
            self.assertEqual(settings.SITE_NAME, "Test Site")
            self.assertEqual(settings.ARCHETYPES, "archetypes")
            self.assertIsNone(settings.CONTENT)
            self.assertEqual(settings.THEME, "default")
            self.assertIsNone(settings.OUTPUT)
            self.assertIsNone(settings.TIMEZONE)
            self.assertEqual(settings.SYSTEM_ASSETS, [])
            self.assertFalse(settings.DRAFT_MODE)
            self.assertTrue(settings.CACHE_REBUILDS)


class TestExampleProject(test_utils.SiteTestMixin, TestCase):
    site_name = "demo"

    def test_settings(self):
        args = test_utils.Args(project=self.mocksite.root)
        cmd = SiteCommand(args)
        settings = cmd.settings
        self.assertEqual(settings.PROJECT_ROOT, self.mocksite.root)
        self.assertEqual(settings.SITE_ROOT, "/")
        # self.assertEqual(settings.SITE_NAME, "Example web site")
        self.assertIsNone(settings.SITE_NAME)
        self.assertEqual(settings.ARCHETYPES, "archetypes")
        self.assertEqual(settings.CONTENT, "content")
        self.assertEqual(settings.THEME, "default")
        self.assertEqual(settings.OUTPUT, "web")
        self.assertEqual(settings.TIMEZONE, "Europe/Rome")
        self.assertEqual(settings.SYSTEM_ASSETS, [])
        self.assertFalse(settings.DRAFT_MODE)
        self.assertTrue(settings.CACHE_REBUILDS)

    def test_default_name(self):
        output = os.path.join(self.build_root, "index.html")
        self.assertTrue(os.path.exists(output))

        with open(output, encoding="utf8") as fd:
            content = fd.read()

        # FIXME: this does not make sense anymore, since title is not defined
        # by settings anymore
        self.assertIn("Example web site", content)



from __future__ import annotations

import datetime
from unittest import TestCase

from . import utils as test_utils


class TestTags(test_utils.MockSiteTestMixin, TestCase):
    def test_site(self):
        """
        Test simply assigning pages to taxonomies
        """
        files = {
            "taxonomies/tags.taxonomy": "---\n",
            "page1.md": {
                "date": str(datetime.datetime(2016, 1, 1)),
                "tags": ["a", "b"],
            },
        }
        with self.site(files) as mocksite:
            tags, page1 = mocksite.page("taxonomies/tags", "page1")
            self.assertEqual(tags.categories["a"].pages, [page1])
            self.assertEqual(tags.categories["b"].pages, [page1])

    def test_load(self):
        """
        Test simply assigning pages to taxonomies
        """
        files = {
            "taxonomies/tags.taxonomy": "---\n",
            "page.md": """---
tags: [cat]
---
# Title
""",
        }
        with self.site(files) as mocksite:
            tags, page = mocksite.page("taxonomies/tags", "page")
            cat = tags.categories["cat"]
            self.assertEqual(cat.pages, [page])
            self.assertEqual(page.tags, [cat])

    def test_field_validation(self):
        """
        Test the various ways of providing tags
        """
        files = {
            "tags.taxonomy": "---\n",
            "single.md": {"tags": "cat"},
            "multi.md": {"tags": ["cat", "dog"]},
            "none1.md": {"tags": ""},
            "none2.md": {"tags": []},
            "none3.md": {"tags": None},
        }
        with self.site(files) as mocksite:
            single, multi, none1, none2, none3 = mocksite.page(
                "single", "multi", "none1", "none2", "none3"
            )
            cat, dog = mocksite.page("tags/cat", "tags/dog")
            self.assertEqual(single.tags, [cat])
            self.assertEqual(multi.tags, [cat, dog])
            self.assertEqual(none1.tags, [])
            self.assertEqual(none2.tags, [])
            self.assertEqual(none3.tags, [])

    def test_enrico_tags(self):
        """
        Test tags.taxonomy from my own site
        """
        files = {
            "tags.taxonomy": {
                "title": "All tags",
                "description": "Index of all tags in the site.",
                "category": {
                    "template_title": "Latest posts for tag <strong>{{page.name}}</strong>",
                    "template_description": "Most recent posts with tag <strong>{{page.name}}</strong>",
                    "syndication": {
                        "template_title": "{{page.meta.site_name}}: posts with tag {{page.meta.index.name}}",
                        "template_description": "{{page.meta.site_name}}: most recent posts"
                        " with tag {{page.meta.index.name}}",
                    },
                },
                "archive": {
                    "template_title": "Archive of posts for tag <strong>{{page.created_from.name}}</strong>",
                    "template_description": "Archive of all posts with tag <strong>{{page.created_from.name}}</strong>",
                },
            },
            "page.md": {
                "tags": ["test", "test1"],
                "title": "Page",
            },
        }
        with self.site(files) as mocksite:
            mocksite.assertPagePaths(
                (
                    "",
                    "page",
                    "tags",
                    "tags/test",
                    "tags/test/index.rss",
                    "tags/test/index.atom",
                    "tags/test/archive",
                    "tags/test1",
                    "tags/test1/index.rss",
                    "tags/test1/index.atom",
                    "tags/test1/archive",
                )
            )

            page, tags = mocksite.page("page", "tags")
            test, rss, atom, archive = mocksite.page(
                "tags/test",
                "tags/test/index.rss",
                "tags/test/index.atom",
                "tags/test/archive",
            )
            test1, rss1, atom1, archive1 = mocksite.page(
                "tags/test1",
                "tags/test1/index.rss",
                "tags/test1/index.atom",
                "tags/test1/archive",
            )

            self.assertEqual(page.title, "Page")
            self.assertIsNone(page.description)
            self.assertEqual(page.tags, [test, test1])
            self.assertEqual(page.related, {})

            self.assertEqual(test.title, "Latest posts for tag <strong>test</strong>")
            self.assertEqual(
                test.description, "Most recent posts with tag <strong>test</strong>"
            )
            self.assertEqual(test.name, "test")
            self.assertEqual(test.taxonomy, tags)
            self.assertEqual(
                test.related,
                {
                    "rss_feed": rss,
                    "atom_feed": atom,
                    "archive": archive,
                },
            )

            self.assertEqual(rss.title, "Test site: posts with tag test")
            self.assertEqual(
                rss.description, "Test site: most recent posts with tag test"
            )
            self.assertEqual(rss.related, {})
            self.assertEqual(rss.index, test)

            self.assertEqual(
                archive.title, "Archive of posts for tag <strong>test</strong>"
            )
            self.assertEqual(
                archive.description,
                "Archive of all posts with tag <strong>test</strong>",
            )
            self.assertEqual(
                archive.related,
                {
                    "rss_feed": rss,
                    "atom_feed": atom,
                },
            )
            self.assertEqual(archive.index, test)

            self.assertEqual(rss1.title, "Test site: posts with tag test1")
            self.assertEqual(
                rss1.description, "Test site: most recent posts with tag test1"
            )
            self.assertEqual(rss1.related, {})
            self.assertEqual(rss1.index, test1)

            self.assertEqual(
                archive1.title, "Archive of posts for tag <strong>test1</strong>"
            )
            self.assertEqual(
                archive1.description,
                "Archive of all posts with tag <strong>test1</strong>",
            )
            self.assertEqual(
                archive1.related,
                {
                    "rss_feed": rss1,
                    "atom_feed": atom1,
                },
            )
            self.assertEqual(archive1.index, test1)



from __future__ import annotations

import importlib
import logging
import sys
import types
from collections.abc import Sequence
from typing import Any

log = logging.getLogger("settings")


class Settings:
    # `ssite` command being run
    BUILD_COMMAND: str

    # Root directory used to resolve relative path in settings
    # Default if None: the directory where the settings file is found
    PROJECT_ROOT: str | None

    # Base URL for the site, used to generate absolute URLs
    SITE_URL: str

    # Root directory of the site in the URLs we generate.
    #
    # If you are publishing the site at /prefix instead of root of the domain,
    # override this with /prefix
    SITE_ROOT: str

    # Default site name. If None, use the title of the toplevel index
    SITE_NAME: str | None

    # Default author of the site
    SITE_AUTHOR: str | None

    # Directory with "archetypes" (templates used by ssite new)
    # If None, archetypes are not used by ssite new
    ARCHETYPES: str

    # Directory with the source content of the site
    # Default if None: PROJECT_ROOT
    CONTENT: str | None

    # Directories where themes are looked for
    THEME_PATHS: Sequence[str]

    # Theme used to render the site.
    # For compatibility, if it is a sequence of strings, it is treated as a list of
    # full paths to theme directories to try in order
    THEME: str | Sequence[str]

    # Directory where the static site will be written by build
    # If None, require providing it explicitly to build
    OUTPUT: str | None

    # Time zone used for timestamps on the site
    # (NONE defaults to the system configured timezone)
    TIMEZONE: str | None

    # Editor used to edit new pages
    EDITOR: str

    # Command used to run the editor, as passed to subprocess.check_command.
    # Each list element is expanded with string.format. All settings are available
    # for expansion, and {name} is the absolute path of the file to edit.
    EDIT_COMMAND: Sequence[str]

    # extensions for python-markdown and their config used for this site
    MARKDOWN_EXTENSIONS: list[str]
    MARKDOWN_EXTENSION_CONFIGS: dict[str, Any]

    # List of asset directories included from /usr/share/javascript
    SYSTEM_ASSETS: Sequence[str]

    # If true, do not ignore pages with dates in the future
    DRAFT_MODE: bool

    # If True, store cached data to speed up rebuilds
    CACHE_REBUILDS: bool

    # Patterns (glob or regexps) that identify files in content directories that
    # are parsed as jinja2 templates
    JINJA2_PAGES: Sequence[str]

    # Set to false to disable running Jinja2 in a sandboxed environnment.
    # If you trust your site sources, it renders noticeably faster.
    JINJA2_SANDBOXED: bool

    # Languages used to build the site
    # For now, only the first one is used, and only its locale is used.
    LANGUAGES: Sequence[dict[str, Any]]

    # Path where theme static assets will be placed in built site
    # Override with "" to merge them with the rest of the contents
    STATIC_PATH: str

    def __init__(self, default_settings: str = "staticsite.global_settings") -> None:
        if default_settings is not None:
            self.add_module(importlib.import_module(default_settings))

    def as_dict(self) -> dict[str, Any]:
        res = {}
        for setting in dir(self):
            if setting.isupper():
                res[setting] = getattr(self, setting)
        return res

    def add_module(self, mod: types.ModuleType) -> None:
        """
        Add uppercase settings from mod into this module
        """
        for setting in dir(mod):
            if setting.isupper():
                setattr(self, setting, getattr(mod, setting))

    def load(self, pathname: str) -> None:
        """
        Load settings from a python file, importing only uppercase symbols
        """
        orig_dwb = sys.dont_write_bytecode
        try:
            sys.dont_write_bytecode = True
            # http://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path

            # Seriously, this should not happen in the standard library. You do not
            # break stable APIs. You can extend them but not break them. And
            # especially, you do not break stable APIs and then complain that people
            # stick to 2.7 until its death, and probably after.
            import importlib.util

            spec = importlib.util.spec_from_file_location(
                "staticsite.settings", pathname
            )
            user_settings = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(user_settings)
        finally:
            sys.dont_write_bytecode = orig_dwb

        self.add_module(user_settings)



import io
from unittest import SkipTest, TestCase

from staticsite.utils import yaml_codec

yaml_sample = """---
key: val
bool: true
"""

yaml_sample_sorted = """---
bool: true
key: val
"""

yaml_sample_parsed = {
    "key": "val",
    "bool": True,
}


class YamlTestMixin:
    def test_loads(self):
        self.assertEqual(self.loads(self.sample), self.sample_parsed)
        self.assertEqual(self.loads(self.sample.encode()), self.sample_parsed)

    def test_load(self):
        with io.StringIO(self.sample) as fd:
            self.assertEqual(self.load(fd), self.sample_parsed)

    def test_dumps(self):
        self.assertEqual(self.dumps(self.sample_parsed), self.sample)

    def test_dump(self):
        with io.StringIO() as fd:
            self.dump(self.sample_parsed, fd)
            self.assertEqual(fd.getvalue(), self.sample)


class TestRuamel(YamlTestMixin, TestCase):
    def setUp(self):
        if yaml_codec.load_ruamel is None:
            raise SkipTest("ruamel.yaml not available")
        self.loads = yaml_codec.loads_ruamel
        self.load = yaml_codec.load_ruamel
        self.dumps = yaml_codec.dumps_ruamel
        self.dump = yaml_codec.dump_ruamel
        self.sample = yaml_sample
        self.sample_parsed = yaml_sample_parsed


class TestPyYAML(YamlTestMixin, TestCase):
    def setUp(self):
        if yaml_codec.load_pyyaml is None:
            raise SkipTest("PyYAML not available")
        self.loads = yaml_codec.loads_pyyaml
        self.load = yaml_codec.load_pyyaml
        self.dumps = yaml_codec.dumps_pyyaml
        self.dump = yaml_codec.dump_pyyaml
        self.sample = yaml_sample_sorted
        self.sample_parsed = yaml_sample_parsed



from __future__ import annotations

import logging
from collections import Counter
from collections.abc import Iterable, Sequence
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from staticsite.page import Page

log = logging.getLogger("links")


class Link:
    """
    All known information about one external link
    """

    def __init__(self, info: dict[str, Any], page: Page | None = None):
        self.page = page
        self.url: str = info["url"]
        self.title: str = info["title"]
        self.tags: set[str] = set(info.get("tags", ()))
        self.archive: str | None = info.get("archive")
        self.related: Sequence[str] = info.get("related", ())
        self.abstract: str | None = info.get("abstract")

    def as_dict(self) -> dict[str, Any]:
        res: dict[str, Any] = {"url": self.url}
        if self.title:
            res["title"] = self.title
        if self.tags:
            res["tags"] = list(self.tags)
        if self.archive:
            res["archive"] = self.archive
        if self.related:
            res["related"] = self.related
        if self.abstract:
            res["abstract"] = self.abstract
        return res


class LinkCollection:
    """
    Collection of links
    """

    def __init__(self, links: dict[str, Link] | None = None):
        self.links: dict[str, Link] = links if links is not None else {}

        # Compute count of link and tag cardinalities
        self.card: Counter[str] = Counter()
        for link in self.links.values():
            for tag in link.tags:
                self.card[tag] += 1

    def __iter__(self) -> Iterable[Link]:
        return self.links.values().__iter__()

    def __len__(self) -> int:
        return len(self.links)

    def __bool__(self) -> bool:
        return bool(self.links)

    def get(self, url: str) -> Link | None:
        return self.links.get(url)

    def tags_and_cards(self) -> list[tuple[str, int]]:
        """
        Return a list of (tag, card) for all tags in the collection
        """
        res = list(
            (tag, card)
            for tag, card in self.card.most_common()
            if card < len(self.links)
        )
        res.sort(key=lambda x: (-x[1], x[0]))
        return res

    def append(self, link: Link) -> None:
        self.links[link.url] = link
        for tag in link.tags:
            self.card[tag] += 1

    def merge(self, coll: LinkCollection) -> None:
        self.links.update(coll.links)
        self.card += coll.card

    def make_groups(self) -> list[tuple[str | None, LinkCollection]]:
        """
        Compute groups from links.

        Return a dict with the grouper tag mapped to the grouped links
        """
        seen: set[str] = set()
        groups: list[tuple[str | None, LinkCollection]] = []
        for tag, card in self.card.most_common():
            if card < 4:
                break
            selected = {
                link.url: link for link in self.links.values() if tag in link.tags
            }
            if len(selected) == len(self.links):
                continue
            seen.update(link.url for link in selected.values())
            groups.append((tag, LinkCollection(selected)))

        groups.sort()

        others = {
            link.url: link for link in self.links.values() if link.url not in seen
        }
        if others:
            # If all elements in other share a tag, use that instead of None
            other_tag: str | None = None
            common_tags: set[str] = set.intersection(
                *(link.tags for link in others.values())
            )
            if common_tags:
                other_tag = sorted(common_tags)[0]
            groups.append((other_tag, LinkCollection(others)))

        return groups



from __future__ import annotations

from __future__ import annotations
from typing import ContextManager, Optional, TypeVar, Union, overload

T = TypeVar("T")


class _Database:
    ...


class Environment:
    def begin(
            self,
            db: Optional[_Database] = None,
            parent: Optional[Transaction] = None,
            write: bool = False,
            buffers: bool = False) -> ContextManager[Transaction]:
        ...

    @overload
    def get(self, key: bytes, default: bytes, db: Optional[_Database] = None) -> bytes:
        ...

    @overload
    def get(self, key: bytes, default: T, db: Optional[_Database] = None) -> Union[bytes, T]:
        ...

    def put(
            self,
            key: str, value: bytes,
            dupdata: bool = True,
            overwrite: bool = True,
            append: bool = False,
            db: Optional[_Database] = None) -> None:
        ...


class Transaction:
    @overload
    def get(self, key: bytes, default: bytes) -> bytes:
        ...

    @overload
    def get(self, key: bytes, default: T) -> Union[bytes, T]:
        ...

    def put(
            self,
            key: bytes, value: bytes,
            dupdata: bool = True,
            overwrite: bool = True,
            append: bool = False,
            db: Optional[_Database] = None) -> None:
        ...


def open(
        path: str,
        map_size: int = 10485760,
        subdir: bool = True,
        readonly: bool = False,
        metasync: bool = True,
        sync: bool = True,
        map_async: bool = False,
        mode: int = 493,
        create: bool = True,
        readahead: bool = True,
        writemap: bool = False,
        meminit: bool = True,
        max_readers: int = 126,
        max_dbs: int = 0,
        max_spare_txns: int = 1,
        lock: bool = True) -> Environment:
    ...
